#!/usr/bin/env python3
#
#  dicts.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from collections import defaultdict
from itertools import chain


class AttrDict(dict):
	"""
	Subclass of dict with attribute access to keys
	"""

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.__dict__ = self


class defaultkeydict(defaultdict):
	"""
	defaultdict where default_factory should have the signature key -> value

	Examples
	--------
	>>> d = defaultkeydict(lambda k: f'[{k}]' , {'a': '[a]', 'b': '[B]'})
	>>> d['a']
	'[a]'
	>>> d['b']
	'[B]'
	>>> d['c']
	'[c]'
	"""

	def __missing__(self, key):
		if self.default_factory is None:
			raise KeyError(f"Missing key: {key}")
		else:
			self[key] = self.default_factory(key)
		return self[key]


def _imul(d1, d2):
	if hasattr(d2, 'keys'):
		for k in set(chain(d1.keys(), d2.keys())):
			d1[k] = d1[k] * d2[k]
	else:
		for k in d1:
			d1[k] *= d2


def _itruediv(d1, d2):
	if hasattr(d2, 'keys'):
		for k in set(chain(d1.keys(), d2.keys())):
			d1[k] = d1[k] / d2[k]
	else:
		for k in d1:
			d1[k] /= d2


class ArithmeticDict(defaultdict):
	"""
		A dictionary which supports arithmetics

	Subclassed from defaultdict, with support for addition, subtraction,
	multiplication and division. If other term/factor has a :meth:`keys` method
	the arithmetics are performed on a key per key basis. If :meth:`keys` is
	missing, the operation is broadcasted onto all values.
	Nonexisting keys are interpreted to signal a zero.

	Notes
	-----
	``__eq__`` ignores values equal to ``self.default_factory()``

	Examples
	--------
	>>> d1 = ArithmeticDict(float, {'a': 2.0, 'b': 3.0})
	>>> d2 = ArithmeticDict(float, {'b': 5.0, 'c': 7.0})
	>>> (d1 + d2) == {'a': 2., 'b': 8., 'c': 7., 'd': 0.}
	True
	>>> (d1 * d1) == {'a': 4.0, 'b': 9.0, 'z': 0}
	True
	>>> (d1 * d2) == {'b': 15}
	True
	>>> d1*2 == {'a': 4, 'b': 6}
	True
	>>> (d1 / {'a': 2, 'b': 11})['b'] == 3./11
	True
	>>> d2/3 == {'b': 5./3, 'c': 7./3}
	True

	"""

	def copy(self):
		return self.__class__(self.default_factory, self.items())

	def __iadd__(self, other):
		try:
			for k, v in other.items():
				self[k] += v
		except AttributeError:
			for k in self:
				self[k] += other
		return self

	def __isub__(self, other):
		try:
			for k, v in other.items():
				self[k] -= v
		except AttributeError:
			for k in self:
				self[k] -= other
		return self

	def __add__(self, other):
		a = self.copy()
		a += other
		return a

	def __sub__(self, other):
		a = self.copy()
		a -= other
		return a

	def __radd__(self, other):
		return self + other

	def __rsub__(self, other):
		return -1 * self + other

	def __imul__(self, other):
		_imul(self, other)
		return self

	def __mul__(self, other):
		a = self.copy()
		a *= other
		return a

	def __rmul__(self, other):
		return self * other

	def __itruediv__(self, other):
		_itruediv(self, other)
		return self

	def __truediv__(self, other):
		a = self.copy()
		a /= other
		return a

	def __rtruediv__(self, other):
		"""
		other / self """
		return self.__class__(
				self.default_factory,
				{k: other / v for k, v in self.items()})

	def __ifloordiv__(self, other):
		if hasattr(other, 'keys'):
			for k in set(chain(self.keys(), other.keys())):
				self[k] = self[k] // other[k]
		else:
			for k in self:
				self[k] //= other
		return self

	def __floordiv__(self, other):
		a = self.copy()
		a //= other
		return a

	def __rfloordiv__(self, other):
		"""
		other // self """
		return self.__class__(
				self.default_factory,
				{k: other // v for k, v in self.items()})

	def __repr__(self):
		return f"{self.__class__.__name__}({repr(self.default_factory)}, {dict(self)})"

	def _element_eq(self, a, b):
		return a == b

	def _discrepancy(self, other, cb):
		default = self.default_factory()
		_self = self.copy()  # getitem is not idempotent on defaultdict
		_other = other.copy()
		try:
			for k in set(chain(_self.keys(), _other.keys())):
				if not cb(_self[k], _other.get(k, default)):
					return False
			return True
		except TypeError:
			return False

	def __eq__(self, other):
		return self._discrepancy(other, self._element_eq)

	def isclose(self, other, rtol=1e-12, atol=None):
		def _isclose(a, b):
			lim = abs(rtol * b)
			if atol is not None:
				lim += atol
			return abs(a - b) <= lim
		return self._discrepancy(other, _isclose)

	def all_non_negative(self):
		for v in self.values():
			if v < v * 0:
				return False
		return True



# -*- coding: utf-8 -*-
"""
test_pandas
~~~~~~~~~~~

Test optional pandas functionality.

"""


# 3rd party
import pytest
pandas = pytest.importorskip('pandas')

# this package
from chemistry_tools.pubchem.compound import compounds_to_frame, Compound
from chemistry_tools.pubchem.lookup import get_compounds
from chemistry_tools.pubchem.properties import get_properties


@pytest.mark.skip("TODO")
@pytest.mark.flaky(reruns=5, reruns_delay=10)
def test_compounds_dataframe():
	df = get_compounds('C20H41Br', 'formula', as_dataframe=True)
	assert df.ndim == 2
	assert df.index.names == ['CID']
	assert len(df.index) > 5
	columns = df.columns.values.tolist()
	assert 'atom_stereo_count' in columns
	assert 'atoms' in columns
	assert 'canonical_smiles' in columns
	assert 'exact_mass' in columns


def test_properties_dataframe():
	df = get_properties('1,2,3,4', ['IsomericSMILES', 'XLogP', 'InChIKey'], 'cid', as_dataframe=True)
	assert df.ndim == 2
	assert df.index.names == ['CID']
	assert len(df.index) == 4
	assert df.columns.values.tolist() == ['IsomericSMILES', 'InChIKey', 'XLogP']


def test_compound_series():
	s = Compound.from_cid(241).to_series()
	assert isinstance(s, pandas.Series)


def test_compound_to_frame():
	s = compounds_to_frame(Compound.from_cid(241))
	assert isinstance(s, pandas.DataFrame)



#!/usr/bin/env python
#
#  bond.py
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#


# 3rd party
from domdf_python_tools.enums import IntEnum

# this package
from chemistry_tools.pubchem.errors import ResponseParseError


class BondType(IntEnum):
	SINGLE = 1
	DOUBLE = 2
	TRIPLE = 3
	QUADRUPLE = 4
	DATIVE = 5
	COMPLEX = 6
	IONIC = 7
	UNKNOWN = 255


class Bond:
	"""
	Class to represent a bond between two atoms in a :class:`~pubchempy.Compound`.
	"""

	def __init__(self, aid1, aid2, order=BondType.SINGLE, style=None):
		"""
		Initialize with begin and end atom IDs, bond order and bond style.

		:param aid1: ID of the begin atom of this bond
		:type aid1: int
		:param aid2: ID of the end atom of this bond
		:type aid2: int
		:param order: Bond order.
		:type order: :class:`chemistry_tools.pubchem.bond.BondType`
		:param style: Bond style annotation.
		:type style:
		"""

		self.aid1 = aid1
		self.aid2 = aid2
		self.order = BondType(order)
		self.style = style

	def __repr__(self):
		return 'Bond({}, {}, {})'.format(self.aid1, self.aid2, self.order)

	def __eq__(self, other):
		return (
				isinstance(other, type(self))
				and self.aid1 == other.aid1
				and self.aid2 == other.aid2
				and self.order == other.order
				and self.style == other.style
				)

	def to_dict(self):
		"""
		Return a dictionary containing Bond data.
		"""

		data = {'aid1': self.aid1, 'aid2': self.aid2, 'order': self.order}

		if self.style is not None:
			data['style'] = self.style

		return data



def parse_bonds(bonds_dict, coords_dict=None):
	"""

	:param bonds_dict:
	:type bonds_dict: dict
	:param coords_dict:
	:type coords_dict: dict or None

	:return:
	:rtype: dict
	"""

	bonds = {}

	if not bonds_dict:
		return bonds

	# Create bonds
	aid1s = bonds_dict['aid1']
	aid2s = bonds_dict['aid2']
	orders = bonds_dict['order']

	if not len(aid1s) == len(aid2s) == len(orders):
		raise ResponseParseError('Error parsing bonds')

	for aid1, aid2, order in zip(aid1s, aid2s, orders):
		bonds[frozenset((aid1, aid2))] = Bond(aid1=aid1, aid2=aid2, order=order)

	# Add styles
	if coords_dict and 'style' in coords_dict[0]['conformers'][0]:
		aid1s = coords_dict[0]['conformers'][0]['style']['aid1']
		aid2s = coords_dict[0]['conformers'][0]['style']['aid2']
		styles = coords_dict[0]['conformers'][0]['style']['annotation']
		for aid1, aid2, style in zip(aid1s, aid2s, styles):
			bonds[frozenset((aid1, aid2))].style = style

	return bonds



# -*- coding: utf-8 -*-
"""
test_Atom
~~~~~~~~~~~~~

Test Atom class

"""

import pytest

from chemistry_tools.pubchem.atom import Atom


@pytest.fixture(scope='module')
def a1():
	return Atom(1234, 6, 7, 8, charge=-1)


def test_atom(a1):
	assert a1.__repr__() == "Atom(1234, C)"
	assert a1.element == "C"
	assert isinstance(a1.element, str)
	assert isinstance(a1.to_dict(), dict)
	assert a1.to_dict()["number"] == 6
	assert a1.aid == 1234
	assert a1.x == 7
	assert a1.y == 8
	assert a1.z is None


def test_coordinates(a1):
	a1.set_coordinates(7, 8, 9)
	assert a1.coordinate_type == "3d"

	a1.set_coordinates(7, 8)
	assert a1.coordinate_type == "2d"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  synonyms.py
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


# this package
from chemistry_tools.pubchem.enums import PubChemNamespace
from chemistry_tools.pubchem.pug_rest import _do_rest_get


class Synonyms(list):
	def __init__(self, initlist):
		super().__init__()

		for val in initlist:
			self.append(str(val))

	def __contains__(self, value):
		for v in self:
			if self.__prep_contains(v) == self.__prep_contains(value):
				return True
		return False

	def append(self, item):
		if item not in self:
			super().append(str(item))

	@staticmethod
	def __prep_contains(val):
		val = val.casefold()

		for remove in ["-", "_", " "]:
			val = val.replace(remove, " ")

		return val


def get_synonyms(identifier, namespace="name"):
	"""
	Returns a list of synonyms for the compound with the given identifier.
	As more than one compound may be identified the results are returned in a list.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:type identifier: str, Sequence[str]
	:param namespace: The type of identifier to look up. Valid values are in :class:`PubChemNamespace`. Default "name"
	:type namespace: PubChemNamespace, optional

	:return: List of dictionaries containing the CID and a list of synonyms for the compounds
	:rtype: List[dict]
	"""

	data = rest_get_synonyms(identifier, namespace)

	results = []

	for compound in data["InformationList"]["Information"]:
		parsed_data = {
				"CID": compound["CID"],
				"synonyms": Synonyms(compound["Synonym"][:20]),
				}

		results.append(parsed_data)

	return results


def rest_get_synonyms(identifier, namespace=PubChemNamespace.name, **kwargs):
	"""
	Get the list of synonyms for the given compound

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:type identifier: str, Sequence[str]
	:param namespace: The type of identifier to look up. Valid values are in :class:`PubChemNamespace`
	:type namespace: PubChemNamespace, optional
	:param **kwargs: Optional arguments that ``json.loads`` takes.
	:raises ValueError: If the response body does not contain valid json.

	:return: Parsed json data
	:rtype: dict
	"""

	return _do_rest_get(namespace, identifier, domain="synonyms").json(**kwargs)



#!/usr/bin/env python3
#
#  test_parser.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import decimal

import pytest
from mathematical.utils import rounders

from chemistry_tools.formulae.html import string_to_html
from chemistry_tools.formulae.latex import string_to_latex
from chemistry_tools.formulae.parser import string_to_composition, mass_from_composition, relative_atomic_masses
from chemistry_tools.formulae.unicode import string_to_unicode


def test_formula_to_composition():
	assert string_to_composition('H2O') == {"H": 2, "O": 1}
	assert string_to_composition('Fe+3') == {0: 3, "Fe": 1}
	assert string_to_composition('Na+1') == {0: 1, "Na": 1}
	assert string_to_composition('Na+') == {0: 1, "Na": 1}
	assert string_to_composition('Cl-') == {0: -1, "Cl": 1}
	assert string_to_composition('NaCl') == {"Na": 1, "Cl": 1}
	assert string_to_composition('NaCl(s)') == {"Na": 1, "Cl": 1}
	assert string_to_composition('Fe(SCN)2+') == {
			0: 1, "C": 2, "N": 2, "S": 2, "Fe": 1}
	assert string_to_composition('Fe(SCN)2+1') == {
			0: 1, "C": 2, "N": 2, "S": 2, "Fe": 1}
	assert string_to_composition('((H2O)2OH)12') == {"H": 60, "O": 36}

	# Special case: solvated electron:
	assert string_to_composition('e-') == {0: -1}
	assert string_to_composition('e-1') == {0: -1}
	assert string_to_composition('e-(aq)') == {0: -1}
	assert string_to_composition('SO4-2(aq)') == {0: -2, "O": 4, "S": 1}

	# prefixes and suffixes
	assert string_to_composition('.NO2(g)') == {"N": 1, "O": 2}
	assert string_to_composition('.NH2') == {"H": 2, "N": 1}
	assert string_to_composition('ONOOH') == {"H": 1, "N": 1, "O": 3}
	assert string_to_composition('.ONOO') == {"N": 1, "O": 3}
	assert string_to_composition('.NO3-2') == {0: -2, "N": 1, "O": 3}

	with pytest.raises(ValueError):
		string_to_composition('F-F')

	# TODO: parse greek prefixes
	# assert string_to_composition('alpha-FeOOH(s)') == {"H": 1, "O": 2, "Fe": 1}
	# assert string_to_composition('epsilon-Zn(OH)2(s)') == {"H": 2, "O": 2, "Zn": 1}

	assert string_to_composition('Na2CO3.7H2O(s)') == {"Na": 2, "C": 1, "O": 10, "H": 14}


def test_formula_to_latex():
	assert string_to_latex('H2O') == 'H_{2}O'
	assert string_to_latex('C6H6+') == 'C_{6}H_{6}^{+}'
	assert string_to_latex('Fe(CN)6-3') == 'Fe(CN)_{6}^{3-}'
	assert string_to_latex('C18H38+2') == 'C_{18}H_{38}^{2+}'
	assert string_to_latex('((H2O)2OH)12') == '((H_{2}O)_{2}OH)_{12}'
	assert string_to_latex('NaCl') == 'NaCl'
	assert string_to_latex('NaCl(s)') == 'NaCl(s)'
	assert string_to_latex('e-(aq)') == 'e^{-}(aq)'
	assert string_to_latex('Ca+2(aq)') == 'Ca^{2+}(aq)'
	assert string_to_latex('.NO2(g)') == r'^\bullet NO_{2}(g)'
	assert string_to_latex('.NH2') == r'^\bullet NH_{2}'
	assert string_to_latex('ONOOH') == 'ONOOH'
	assert string_to_latex('.ONOO') == r'^\bullet ONOO'
	assert string_to_latex('.NO3-2') == r'^\bullet NO_{3}^{2-}'
	assert string_to_latex('alpha-FeOOH(s)') == r'\alpha-FeOOH(s)'
	assert string_to_latex('epsilon-Zn(OH)2(s)') == (
			r'\varepsilon-Zn(OH)_{2}(s)')
	assert string_to_latex('Na2CO3.7H2O(s)') == r'Na_{2}CO_{3}\cdot 7H_{2}O(s)'
	assert string_to_latex('Na2CO3.1H2O(s)') == r'Na_{2}CO_{3}\cdot H_{2}O(s)'


def test_formula_to_unicoce():
	assert string_to_unicode('NH4+') == 'NH₄⁺'
	assert string_to_unicode('H2O') == 'H₂O'
	assert string_to_unicode('C6H6+') == 'C₆H₆⁺'
	assert string_to_unicode('Fe(CN)6-3') == 'Fe(CN)₆³⁻'
	assert string_to_unicode('C18H38+2') == 'C₁₈H₃₈²⁺'
	assert string_to_unicode('((H2O)2OH)12') == '((H₂O)₂OH)₁₂'
	assert string_to_unicode('NaCl') == 'NaCl'
	assert string_to_unicode('NaCl(s)') == 'NaCl(s)'
	assert string_to_unicode('e-(aq)') == 'e⁻(aq)'
	assert string_to_unicode('Ca+2(aq)') == 'Ca²⁺(aq)'
	assert string_to_unicode('.NO2(g)') == '⋅NO₂(g)'
	assert string_to_unicode('.NH2') == '⋅NH₂'
	assert string_to_unicode('ONOOH') == 'ONOOH'
	assert string_to_unicode('.ONOO') == '⋅ONOO'
	assert string_to_unicode('.NO3-2') == '⋅NO₃²⁻'
	assert string_to_unicode('alpha-FeOOH(s)') == 'α-FeOOH(s)'
	assert string_to_unicode('epsilon-Zn(OH)2(s)') == 'ε-Zn(OH)₂(s)'
	assert string_to_unicode('Na2CO3.7H2O(s)') == 'Na₂CO₃·7H₂O(s)'
	assert string_to_unicode('Na2CO3.1H2O(s)') == 'Na₂CO₃·H₂O(s)'


def test_formula_to_html():
	assert string_to_html('H2O') == 'H<sub>2</sub>O'
	assert string_to_html('C6H6+') == 'C<sub>6</sub>H<sub>6</sub><sup>+</sup>'
	assert string_to_html('Fe(CN)6-3') == 'Fe(CN)<sub>6</sub><sup>3-</sup>'
	assert string_to_html('C18H38+2') == 'C<sub>18</sub>H<sub>38</sub><sup>2+</sup>'
	assert string_to_html('((H2O)2OH)12') == '((H<sub>2</sub>O)<sub>2</sub>OH)<sub>12</sub>'
	assert string_to_html('NaCl') == 'NaCl'
	assert string_to_html('NaCl(s)') == 'NaCl(s)'
	assert string_to_html('e-(aq)') == 'e<sup>-</sup>(aq)'
	assert string_to_html('Ca+2(aq)') == 'Ca<sup>2+</sup>(aq)'
	assert string_to_html('.NO2(g)') == r'&sdot;NO<sub>2</sub>(g)'
	assert string_to_html('.NH2') == r'&sdot;NH<sub>2</sub>'
	assert string_to_html('ONOOH') == 'ONOOH'
	assert string_to_html('.ONOO') == r'&sdot;ONOO'
	assert string_to_html('.NO3-2') == r'&sdot;NO<sub>3</sub><sup>2-</sup>'
	assert string_to_html('alpha-FeOOH(s)') == r'&alpha;-FeOOH(s)'
	assert string_to_html('epsilon-Zn(OH)2(s)') == (
			r'&epsilon;-Zn(OH)<sub>2</sub>(s)')
	assert string_to_html('Na2CO3.7H2O(s)') == 'Na<sub>2</sub>CO<sub>3</sub>&sdot;7H<sub>2</sub>O(s)'
	assert string_to_html('Na2CO3.1H2O(s)') == 'Na<sub>2</sub>CO<sub>3</sub>&sdot;H<sub>2</sub>O(s)'


def test_mass_from_composition():
	mass1 = mass_from_composition({11: 1, 9: 1})
	assert rounders(mass1, "0.000000") == decimal.Decimal("41.988172")

	mass2 = mass_from_composition({"Na": 1, "F": 1})
	assert mass1 == mass2
	assert rounders(mass2, "0.000000") == decimal.Decimal("41.988172")


def test_relative_atomic_masses():
	assert rounders(relative_atomic_masses[0], "0.0000") == decimal.Decimal("1.0079")


def test_mass_from_composition__formula():
	mass = mass_from_composition(string_to_composition('NaF'))
	assert rounders(mass, "0.000000") == decimal.Decimal("41.988172")

	Fminus = mass_from_composition(string_to_composition('F/-'))
	assert abs(Fminus - 18.998403163 - 5.489e-4) < 1e-7



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  elements.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#


# stdlib
from functools import lru_cache

# 3rd party
from domdf_python_tools import doctools
from domdf_python_tools.bases import Dictable
from memoized_property import memoized_property

# this package
from . import _table


class Element(Dictable):
	"""Chemical element.

	Attributes
	----------
	number : int
		Atomic number.
	symbol : str of length 1 or 2
		Chemical symbol.
	name : str
		Name in English.
	group : int
		Group in periodic table.
	period : int
		Period in periodic table.
	block : int
		Block in periodic table.
	series : int
		Index to chemical series.
	protons : int
		Number of protons.
	neutrons : int
		Number of neutrons in the most abundant naturally occurring stable.
		isotope
	nominalmass : int
		Mass number of the most abundant naturally occurring stable isotope.
	electrons : int
		Number of electrons.
	mass : float
		Relative atomic mass. Ratio of the average mass of atoms
		of the element to 1/12 of the mass of an atom of 12C.
	exactmass : float
		Relative atomic mass calculated from the isotopic composition.
	eleneg : float
		Electronegativity (Pauling scale).
	covrad : float
		Covalent radius in Angstrom.
	atmrad :
		Atomic radius in Angstrom.
	vdwrad : float
		Van der Waals radius in Angstrom.
	tboil : float
		Boiling temperature in K.
	tmelt : float
		Melting temperature in K.
	density : float
		Density at 295K in g/cm3 respectively g/L.
	oxistates : str
		Oxidation states.
	eleaffin : float
		Electron affinity in eV.
	eleconfig : str
		Ground state electron configuration.
	eleconfig_dict : dict
		Ground state electron configuration (shell, subshell): electrons.
	eleshells : int
		Number of electrons per shell.
	ionenergy : tuple
		Ionization energies in eV
	isotopes : dict
		Isotopic composition.
		keys: isotope mass number
		values: Isotope(relative atomic mass, abundance)

	"""

	def __init__(
			self, number, symbol, name,
			group=0, period=0, block='', series=0, mass=0.0, eleneg=0.0,
			eleaffin=0.0, covrad=0.0, atmrad=0.0, vdwrad=0.0, tboil=0.0,
			tmelt=0.0, density=0.0, eleconfig='', oxistates='',
			ionenergy=None, isotopes=None, description='',
			):

		super().__init__()

		self._number = number
		self._symbol = symbol
		self._name = name
		self._electrons = number
		self._protons = number
		self._group = group
		self._period = period
		self._block = block
		self._series = series
		self._mass = mass
		self._eleneg = eleneg
		self._eleaffin = eleaffin
		self._covrad = covrad
		self._atmrad = atmrad
		self._vdwrad = vdwrad
		self._tboil = tboil
		self._tmelt = tmelt
		self._density = density
		self._eleconfig = eleconfig
		self._oxistates = oxistates
		self._description = description

		if ionenergy is None:
			self._ionenergy = tuple()
		else:
			self._ionenergy = ionenergy

		self._isotopes = {}

		if isotopes is not None:
			for massnumber, isotope in isotopes.items():
				if isinstance(isotope, Isotope):
					self._isotopes[int(massnumber)] = isotope
				elif isinstance(isotope, (list, tuple)):
					self._isotopes[int(massnumber)] = Isotope(*isotope, massnumber)

	@property
	def __dict__(self):
		return dict(
				number=self._number,
				symbol=self._symbol,
				name=self._name,
				group=self._group,
				period=self._period,
				block=self._block,
				series=self._series,
				mass=self._mass,
				eleneg=self._eleneg,
				eleaffin=self._eleaffin,
				covrad=self._covrad,
				atmrad=self._atmrad,
				vdwrad=self._vdwrad,
				tboil=self._tboil,
				tmelt=self._tmelt,
				density=self._density,
				eleconfig=self._eleconfig,
				oxistates=self._oxistates,
				ionenergy=self._ionenergy,
				isotopes=self._isotopes,
				description=self._description
				)

	@memoized_property
	def number(self):
		return self._number

	@memoized_property
	def symbol(self):
		return self._symbol

	@memoized_property
	def name(self):
		return self._name

	@memoized_property
	def electrons(self):
		return self._electrons

	@memoized_property
	def protons(self):
		return self._protons

	@memoized_property
	def group(self):
		return self._group

	@memoized_property
	def period(self):
		return self._period

	@memoized_property
	def block(self):
		return self._block

	@memoized_property
	def series(self):
		return self._series

	@memoized_property
	def mass(self):
		return self._mass

	@memoized_property
	def molecular_weight(self):
		return self._mass

	@memoized_property
	def eleneg(self):
		return self._eleneg

	@memoized_property
	def eleaffin(self):
		return self._eleaffin

	@memoized_property
	def covrad(self):
		return self._covrad

	@memoized_property
	def atmrad(self):
		return self._atmrad

	@memoized_property
	def vdwrad(self):
		return self._vdwrad

	@memoized_property
	def tboil(self):
		return self._tboil

	@memoized_property
	def tmelt(self):
		return self._tmelt

	@memoized_property
	def density(self):
		return self._density

	@memoized_property
	def eleconfig(self):
		return self._eleconfig

	@memoized_property
	def oxistates(self):
		return self._oxistates

	@memoized_property
	def ionenergy(self):
		return self._ionenergy

	@memoized_property
	def isotopes(self):
		return self._isotopes

	@memoized_property
	def description(self):
		return self._description

	def __str__(self):
		return self.name

	def __repr__(self):
		ionenergy = []
		for i, j in enumerate(self.ionenergy):
			if i and (i % 5 == 0):
				ionenergy.append(f'\n        {j}')
			else:
				ionenergy.append(f'{j}')
		ionenergy = ', '.join(ionenergy)
		if len(self.ionenergy) > 5:
			ionenergy = f'(\n        {ionenergy},\n    )'
		elif len(self.ionenergy) == 1:
			ionenergy = f'({ionenergy},)'
		else:
			ionenergy = f'({ionenergy})'

		isotopes = []
		for massnum in sorted(self.isotopes):
			iso = self.isotopes[massnum]
			isotopes.append(
					'{0}: Isotope({1}, {2}, {0})'.format(
							massnum, iso.mass, iso.abundance
							)
					)
		isotopes = ',\n        '.join(isotopes)
		if len(self.isotopes) > 1:
			isotopes = f'{{\n        {isotopes},\n    }},'
		else:
			isotopes = f'{{{isotopes}}},'

		return ',\n    '.join((
				f"Element(\n    {self.number}, '{self.symbol}', '{self.name}'",
				f"group={self.group}, period={self.period},"
				f" block='{self.block}', series={self.series}",
				f"mass={self.mass}, eleneg={self.eleneg},"
				f" eleaffin={self.eleaffin}",
				f"covrad={self.covrad}, atmrad={self.atmrad},"
				f" vdwrad={self.vdwrad}",
				f"tboil={self.tboil}, tmelt={self.tmelt}, density={self.density}",
				f"eleconfig='{self.eleconfig}'",
				f"oxistates='{self.oxistates}'",
				f"ionenergy={ionenergy}",
				f"isotopes={isotopes}\n)"
				))

	@memoized_property
	def nominalmass(self):
		"""
		Return mass number of most abundant natural stable isotope.
		"""

		nominalmass = 0
		maxabundance = 0
		for massnum, iso in self.isotopes.items():
			if iso.abundance > maxabundance:
				maxabundance = iso.abundance
				nominalmass = massnum
		return nominalmass

	@memoized_property
	def neutrons(self):
		"""
		Return number neutrons in most abundant natural stable isotope.
		"""

		return self.nominalmass - self.protons

	@memoized_property
	def exactmass(self):
		"""
		Return relative atomic mass calculated from isotopic composition.
		"""

		return sum(iso.mass * iso.abundance for iso in self.isotopes.values())

	@memoized_property
	def eleconfig_dict(self):
		"""
		Return electron configuration as dict.
		"""

		adict = {}
		if self.eleconfig.startswith('['):
			base = self.eleconfig.split(' ', 1)[0][1:-1]
			adict.update(_table.ELEMENTS[base].eleconfig_dict)
		for e in self.eleconfig.split()[bool(adict):]:
			adict[(int(e[0]), e[1])] = int(e[2:]) if len(e) > 2 else 1
		return adict

	@memoized_property
	def eleshells(self):
		"""
		Return number of electrons in shell as tuple.
		"""

		eleshells = [0, 0, 0, 0, 0, 0, 0]
		for key, val in self.eleconfig_dict.items():
			eleshells[key[0] - 1] += val
		return tuple(ele for ele in eleshells if ele)

	def validate(self):
		"""Check consistency of data. Raise Error on failure."""
		assert self.period in _table.PERIODS
		assert self.group in _table.GROUPS
		assert self.block in _table.BLOCKS
		assert self.series in _table.SERIES

		if self.number != self.protons:
			raise ValueError(
					f'{self.symbol} - atomic number must equal proton number'
					)
		if self.protons != sum(self.eleshells):
			raise ValueError(
					f'{self.symbol} - number of protons must equal electrons'
					)
		if len(self.ionenergy) > 1:
			ionev_ = self.ionenergy[0]
			for ionev in self.ionenergy[1:]:
				if ionev <= ionev_:
					raise ValueError(
							f'{self.symbol} - ionenergy not increasing'
							)
				ionev_ = ionev

		mass = 0.0
		frac = 0.0
		for iso in self.isotopes.values():
			mass += iso.abundance * iso.mass
			frac += iso.abundance
		if abs(mass - self.mass) > 0.03:
			raise ValueError(
					f'{self.symbol} - average of isotope masses '
					f'({mass:.4f}) != mass ({self.mass:.4f})'
					)
		if abs(frac - 1.0) > 1e-9:
			raise ValueError(
					f'{self.symbol} - sum of isotope abundances != 1.0'
					)


class Isotope(Dictable):
	"""Isotope massnumber, relative atomic mass, and abundance."""

	def __init__(self, mass=0.0, abundance=1.0, massnumber=0):
		super().__init__()

		self._mass = mass
		self._abundance = abundance
		self._massnumber = massnumber

	@memoized_property
	def mass(self):
		return self._mass

	@memoized_property
	def abundance(self):
		return self._abundance

	@memoized_property
	def massnumber(self):
		return self._massnumber

	def __str__(self):
		return '{}, {:.4f}, {:.6f}%'.format(
				self.massnumber, self.mass, self.abundance * 100
				)

	def __repr__(self):
		return 'Isotope({}, {}, {})'.format(
				repr(self.mass), repr(self.abundance), repr(self.massnumber)
				)

	@property
	def __dict__(self):
		return dict(
				mass=self.mass,
				abundance=self.abundance,
				massnumber=self.massnumber,
				)


# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.


# TODO: make frozen
class Elements:
	"""
	Ordered dict of Elements with lookup by number, symbol, and name.
	"""

	def __init__(self, *elements):
		self._list = []
		self._dict = {}
		for element in elements:
			if element.number > len(self._list) + 1:
				raise ValueError('Elements must be added in order')
			if element.number <= len(self._list):
				self._list[element.number - 1] = element
			else:
				self._list.append(element)
			self._dict[element.number] = element
			self._dict[element.symbol] = element
			self.add_alternate_spelling(element, element.name)

	def __str__(self):
		return f'[{", ".join(ele.symbol for ele in self._list)}]'

	def __repr__(self):
		elements = ',\n    '.join(
				'\n    '.join(line for line in repr(element).splitlines())
				for element in self._list
				)
		elements = f'Elements(\n    {elements},\n)'
		return elements

	def __contains__(self, item):
		return item in self._dict

	def __iter__(self):
		return iter(self._list)

	def __len__(self):
		return len(self._list)

	def __getitem__(self, key):
		if isinstance(key, str):
			try:
				return self._dict[key.casefold()]
			except KeyError:
				return self._dict[key]
		elif isinstance(key, int):
			return self._dict[key]
		elif isinstance(key, float):
			return self._dict[int(key)]
		elif isinstance(key, slice):
			start, stop, step = key.indices(len(self._list))
			return self._list[slice(start - 1, stop - 1, step)]
		else:
			try:
				symbol, isotope = self.split_isotope(key)
				return self._dict[symbol.capitalize()]
			except:
				raise KeyError(f"Unknown key: '{key}'")

	@lru_cache()
	def split_isotope(self, string):
		from chemistry_tools.formulae.formula import _split_isotope
		return _split_isotope(string)

	def add_alternate_spelling(self, element, spelling):
		self._dict[spelling] = element
		self._dict[spelling.lower()] = element
		self._dict[spelling.casefold()] = element

	@memoized_property
	def symbols(self):
		return [element.symbol for element in sorted(self._list, key=lambda e: e.number)]

	@memoized_property
	def names(self):
		return [str(element) for element in sorted(self._list, key=lambda e: e.number)]

	@memoized_property
	def lower_names(self):
		return [str(element).lower() for element in sorted(self._list, key=lambda e: e.number)]


@doctools.append_docstring_from(Element)
class HeavyHydrogen(Element):
	"""
	Subclass of Element to handle the Heavy Hydrogen isotopes Deuterium and Tritium
	"""

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		if self.symbol not in {"D", "T"}:
			raise ValueError("'HeavyHydrogen' can only be used for Deuterium and Tritium")

	@memoized_property
	def nominalmass(self):
		"""
		Return mass number of most abundant natural stable isotope.
		"""

		if self.symbol == "D":
			return 2
		elif self.symbol == "T":
			return 3

	@memoized_property
	def as_isotope(self):
		"""
		Return the isotope in H[X] format

		:rtype: str
		"""

		return f'H[{self.nominalmass}]'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  _elements.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#

__all__ = ("ELEMENTS", "H", "D", "T")

from .actinides import *
from .alkali_metals import *
from .alkaline_earth_metals import *
from .chalcogens import *
from .halogens import *
from .lanthanides import *
from .noble_gases import *
from .pnictogens import *
from .tetrels import *
from .transition_metals import *
from .triels import *
from . import classes
import re

# TODO: nominal mass (massnumber) and monoisotopic mass


H = classes.Element(
		1, 'H', 'Hydrogen',
		group=1, period=1, block='s', series=1,
		mass=1.007941, eleneg=2.2, eleaffin=0.75420375,
		covrad=0.32, atmrad=0.79, vdwrad=1.2,
		tboil=20.28, tmelt=13.81, density=0.084,
		eleconfig='1s',
		oxistates='1*, -1',
		ionenergy=(13.5984,),
		isotopes={
				1: (1.00782503207, 0.999885),
				2: (2.0141017778, 0.000115),
				3: (3.0160492777, 0.0),
				4: (4.02781, 0.0),
				5: (5.03531, 0.0),
				6: (6.04494, 0.0),
				7: (7.05275, 0.0),
				},
		description=(
			"Colourless, odourless gaseous chemical element. Lightest and most "
			"abundant element in the universe. Present in water and in all "
			"organic compounds. Chemically reacts with most elements. Discovered "
			"by Henry Cavendish in 1776."
			)
		)

D = classes.HeavyHydrogen(
		1, 'D', 'Deuterium',
		mass=2.0141017778,
		isotopes={
				1: (1.00782503207, 0.999885),
				2: (2.0141017778, 0.000115),
				3: (3.0160492777, 0.0),
				4: (4.02781, 0.0),
				5: (5.03531, 0.0),
				6: (6.04494, 0.0),
				7: (7.05275, 0.0),
				},
		)

T = classes.HeavyHydrogen(
		1, 'T', 'Tritium',
		mass=3.0160492777,
		isotopes={
				1: (1.00782503207, 0.999885),
				2: (2.0141017778, 0.000115),
				3: (3.0160492777, 0.0),
				4: (4.02781, 0.0),
				5: (5.03531, 0.0),
				6: (6.04494, 0.0),
				7: (7.05275, 0.0),
				},
		)

ELEMENTS = classes.Elements(
		H, He,
		Li, Be, B, C, N, O, F, Ne,
		Na, Mg, Al, Si, P, S, Cl, Ar,
		K, Ca, Sc, Ti, V, Cr, Mn, Fe, Co, Ni, Cu, Zn, Ga, Ge, As, Se, Br, Kr,
		Rb, Sr, Y, Zr, Nb, Mo, Tc, Ru, Rh, Pd, Ag, Cd, In, Sn, Sb, Te, I, Xe,
		Cs, Ba, La, Ce, Pr, Nd, Pm, Sm, Eu, Gd, Tb, Dy, Ho, Er, Tm, Yb, Lu, Hf, Ta, W, Re, Os, Ir, Pt, Au, Hg, Tl, Pb,
		Bi, Po, At, Rn,
		Fr, Ra, Ac, Th, Pa, U, Np, Pu, Am, Cm, Bk, Cf, Es, Fm, Md, No, Lr, Rf, Db, Sg, Bh, Hs, Mt, Ds, Rg, Cn, Nh, Fl,
		Mc, Lv, Ts, Og,
		)

# ELEMENTS = Elements(
# H,                                                                                                                          He,
# Li, Be,                                                                                                 B,  C,  N,  O,  F,  Ne,
# Na, Mg,                                                                                                 Al, Si, P,  S,  Cl, Ar,
# K,  Ca,                                                         Sc, Ti, V,  Cr, Mn, Fe, Co, Ni, Cu, Zn, Ga, Ge, As, Se, Br, Kr,
# Rb, Sr,                                                         Y,  Zr, Nb, Mo, Tc, Ru, Rh, Pd, Ag, Cd, In, Sn, Sb, Te, I,  Xe,
# Cs, Ba, La, Ce, Pr, Nd, Pm, Sm, Eu, Gd, Tb, Dy, Ho, Er, Tm, Yb, Lu, Hf, Ta, W,  Re, Os, Ir, Pt, Au, Hg, Tl, Pb, Bi, Po, At, Rn,
# Fr, Ra, Ac, Th, Pa, U,  Np, Pu, Am, Cm, Bk, Cf, Es, Fm, Md, No, Lr, Rf, Db, Sg, Bh, Hs, Mt, Ds, Rg, Cn, Nh, Fl, Mc, Lv, Ts, Og,
# )


alternate_spellings = {
		"Caesium": "Cesium",  # AmE
		"Aluminium": "Aluminum",  # AmE
		"Sulfur": "Sulphur",  # Not IUPAC and not etymologically correct, but widely used in BrE
		}

for element, spelling in alternate_spellings.items():
	ELEMENTS.add_alternate_spelling(ELEMENTS[element], spelling)

for element in [D, T]:
	ELEMENTS.add_alternate_spelling(element, element.name)
	ELEMENTS._dict[element.symbol] = element
	# ELEMENTS._list.append(element)



# -*- coding: utf-8 -*-
"""
test_requests
~~~~~~~~~~~~~

Test basic requests.

"""


from chemistry_tools.pubchem.pug_rest import _do_rest_get


def test_requests():
	"""
	Test a variety of basic raw requests and ensure they don't return an error code.
	"""

	assert _do_rest_get(identifier='c1ccccc1', namespace='smiles').status_code == 200
	assert _do_rest_get(identifier='coumarin', namespace='name', format_='PNG', png_width=50, png_height=50).status_code == 200


def test_content_type():
	"""
	Test content type header matches desired output format.
	"""

	assert _do_rest_get(identifier=241, namespace="cid", format_='JSON').headers['Content-Type'] == 'application/json'
	assert _do_rest_get(identifier=241, namespace="cid", format_='XML').headers['Content-Type'] == 'application/xml'
	# assert _do_rest_get(identifier=241, namespace="cid", format_='SDF').headers['Content-Type'] == 'chemical/x-mdl-sdfile'
	# assert _do_rest_get(identifier=241, namespace="cid", format_='ASNT').headers['Content-Type'] == 'text/plain'
	assert _do_rest_get(identifier=241, namespace="cid", format_='PNG').headers['Content-Type'] == 'image/png'




#!/usr/bin/env python3
#
#  utils.py
"""
General utilities.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


# stdlib
import types
from collections import namedtuple, OrderedDict
from collections.abc import ItemsView, Mapping
from itertools import product


def identity(x):
	return x


def defaultnamedtuple(typename, field_names, defaults=()):
	"""
	Generates a new subclass of tuple with default values.

	Parameters
	----------
	typename : string
		The name of the class.
	field_names : str or iterable
		An iterable of splitable string.
	defaults : iterable
		Default values for ``field_names``, counting ``[-len(defaults):]``.

	Examples
	--------
	>>> Body = defaultnamedtuple('Body', 'x y z density', (1.0,))
	>>> Body.__doc__
	'Body(x, y, z, density)'
	>>> b = Body(10, z=3, y=5)
	>>> b._asdict() == dict(x=10, y=5, z=3, density=1.0)
	True

	Returns
	-------
	A new tuple subclass named ``typename``
	"""

	Tuple = namedtuple(typename, field_names)
	Tuple.__new__.__defaults__ = (None,) * len(Tuple._fields)
	if isinstance(defaults, Mapping):
		Tuple.__new__.__defaults__ = tuple(Tuple(**defaults))
	else:
		nmissing = len(Tuple._fields) - len(defaults)
		defaults = (None,) * nmissing + tuple(defaults)
		Tuple.__new__.__defaults__ = tuple(Tuple(*defaults))
	return Tuple



# -*- coding: utf-8 -*-
"""
test_Compound
~~~~~~~~~~~~~

Test compound object.

"""

# stdlib
import re
from decimal import Decimal

# 3rd party
import pytest

# this package
from chemistry_tools.pubchem.lookup import get_compounds
from chemistry_tools.pubchem.atom import Atom
from chemistry_tools.pubchem.bond import BondType
from chemistry_tools.pubchem.compound import Compound


@pytest.fixture(scope='module')
def c1():
	"""Compound CID 241."""
	return Compound.from_cid(241)


@pytest.fixture(scope='module')
def c2():
	"""Compound CID 175."""
	return Compound.from_cid(175)


def test_basic(c1):
	"""Test Compound is retrieved and has a record and correct CID."""
	assert c1.cid == 241
	assert repr(c1) == 'Compound(241)'


def test_atoms(c1):
	assert len(c1.atoms) == 12
	print([a.element for a in c1.atoms])
	assert set(a.element for a in c1.atoms) == {'C', 'H'}
	assert set(c1.elements) == {'C', 'H'}


def test_single_atom():
	"""Test Compound when there is a single atom and no bonds."""
	c = Compound.from_cid(259)
	assert c.atoms == [Atom(aid=1, number=35, x=2, y=0, charge=-1)]
	assert c.bonds == []


def test_bonds(c1):
	assert len(c1.bonds) == 12
	assert set(int(b.order) for b in c1.bonds) == {int(BondType.SINGLE), int(BondType.DOUBLE)}


def test_charge(c1):
	assert c1.get_property("Charge") == 0


def test_coordinates(c1):
	for a in c1.atoms:
		assert isinstance(a.x, (float, int))
		assert isinstance(a.y, (float, int))
		assert a.z is None


def test_identifiers(c1):
	# precache properties
	c1.precache()

	assert len(c1.canonical_smiles) > 10
	assert len(c1.get_property("IsomericSMILES")) > 10
	assert c1.smiles == "C1=CC=CC=C1"
	assert c1.get_property("InChI").startswith('InChI=')
	assert re.match(r'^[A-Z]{14}-[A-Z]{10}-[A-Z\d]$', c1.get_property("InChIKey"))
	from chemistry_tools.formulae import Formula
	assert isinstance(c1.get_property("MolecularFormula"), Formula)
	assert isinstance(c1.molecular_formula, Formula)
	assert c1.get_property("MolecularFormula").hill_formula == "C6H6"
	assert c1.molecular_formula.hill_formula == "C6H6"


def test_properties_types(c1):
	# precache properties
	c1.precache()

	assert isinstance(c1.molecular_mass, float)
	assert isinstance(c1.molecular_weight, float)
	assert isinstance(c1.get_property("MolecularWeight"), float)
	assert isinstance(c1.iupac_name, str)
	assert isinstance(c1.systematic_name, str)
	assert isinstance(c1.get_property("XLogP"), float)
	assert isinstance(c1.get_property("ExactMass"), float)
	assert isinstance(c1.get_property("MonoisotopicMass"), float)
	assert isinstance(c1.get_property("TPSA"), (int, float))
	assert isinstance(c1.get_property("Complexity"), float)
	assert isinstance(c1.get_property("HBondDonorCount"), int)
	assert isinstance(c1.get_property("HBondAcceptorCount"), int)
	assert isinstance(c1.get_property("RotatableBondCount"), int)
	assert isinstance(c1.get_property("HeavyAtomCount"), int)
	assert isinstance(c1.get_property("IsotopeAtomCount"), int)
	assert isinstance(c1.get_property("AtomStereoCount"), int)
	assert isinstance(c1.get_property("DefinedAtomStereoCount"), int)
	assert isinstance(c1.get_property("UndefinedAtomStereoCount"), int)
	assert isinstance(c1.get_property("BondStereoCount"), int)
	assert isinstance(c1.get_property("DefinedBondStereoCount"), int)
	assert isinstance(c1.get_property("UndefinedBondStereoCount"), int)
	assert isinstance(c1.get_property("CovalentUnitCount"), int)
	assert isinstance(c1.fingerprint, str)
	# TODO:assert isinstance(c1.hill_formula, text_types)
	assert isinstance(c1.canonicalized, bool) # TODO


def test_properties_values(c1):
	# precache properties
	c1.precache()

	assert c1.molecular_mass == 78.11
	assert c1.molecular_weight == 78.11
	assert c1.molecular_mass == c1.molecular_weight
	assert c1.iupac_name == "benzene"
	assert c1.systematic_name == "benzene"
	assert c1.get_property("XLogP") == 2.1
	assert c1.get_property("ExactMass") == 78.04695
	assert c1.get_property("MonoisotopicMass") == 78.04695
	assert c1.get_property("TPSA") == Decimal(0)
	# assert c1.get_property("Complexity") == 15.5  # TODO: full record has 15.5 but getting just property gives 15
	assert c1.get_property("HBondDonorCount") == Decimal("0")
	assert c1.get_property("HBondAcceptorCount") == Decimal("0")
	assert c1.get_property("RotatableBondCount") == Decimal("0")
	assert c1.get_property("HeavyAtomCount") == Decimal("6")
	assert c1.get_property("IsotopeAtomCount") == Decimal("0")
	assert c1.get_property("AtomStereoCount") == 0
	assert c1.get_property("DefinedAtomStereoCount") == Decimal("0")
	assert c1.get_property("UndefinedAtomStereoCount") == Decimal("0")
	assert c1.get_property("BondStereoCount") == 0
	assert c1.get_property("DefinedBondStereoCount") == Decimal("0")
	assert c1.get_property("UndefinedBondStereoCount") == Decimal("0")
	assert c1.get_property("CovalentUnitCount") == Decimal("1")
	# TODO:assert c1.hill_formula == 'C<sub>6</sub>H<sub>6</sub>'
	assert c1.canonicalized is True


def test_coordinate_type(c1):
	assert c1.coordinate_type == '2d'


def test_compound_equality():
	assert Compound.from_cid(241) == Compound.from_cid(241)
	assert get_compounds('Benzene', 'name')[0], get_compounds('c1ccccc1' == 'smiles')[0]


def test_synonyms(c1):
	assert len(c1.synonyms) > 5
	assert len(c1.synonyms) > 5


def test_compound_dict(c1):
	assert dict(c1)
	assert isinstance(dict(c1), dict)
	assert 'atoms' in dict(c1)
	assert 'bonds' in dict(c1)
	assert dict(c1)['atoms'][1].element


def test_charged_compound(c2):
	assert len(c2.atoms) == 7
	assert c2.atoms[0].charge == -1


def test_fingerprint(c1):
	# CACTVS fingerprint is 881 bits
	assert len(c1.cactvs_fingerprint) == 881
	# Raw fingerprint has 4 byte prefix, 7 bit suffix, and is hex encoded (/4) = 230
	assert len(c1.fingerprint) == (881 + (4 * 8) + 7) / 4

# TODO: Compound.to_series, compounds_to_frame()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2020 Dominic Davis-Foster
#
# Adapted from SpectrumSimilarity.R
# Part of OrgMassSpecR
# Copyright 2011-2017 Nathan Dodder <nathand@sccwrp.org>
# Available under the BSD 2-Clause License


__author__ = "Dominic Davis-Foster"
__copyright__ = "Copyright 2019 Dominic Davis-Foster"

__license__ = "LGPL"
__version__ = "0.1.0"
__email__ = "dominic@davis-foster.co.uk"

import numpy as np
import pandas as pd


def SpectrumSimilarity(
		spec_top, spec_bottom, t=0.25, b=10, top_label=None,
		bottom_label=None, xlim=(50, 1200), x_threshold=0, print_alignment=False,
		print_graphic=True, output_list=False,
		):
	"""

	:param spec_top: Array containing the experimental spectrum's peak list with the m/z values in the
		first column and corresponding intensities in the second
	:type spec_top: numpy.array
	:param spec_bottom: Array containing the reference spectrum's peak list with the m/z values in the
		first column and corresponding intensities in the second
	:type spec_bottom: numpy.array
	:param t: numeric value specifying the tolerance used to align the m/z values of the two spectra.
	:type t: float
	:param b: numeric value specifying the baseline threshold for peak identification. Expressed as a percent of the maximum intensity.
	:type b: float
	:param top_label: string to label the top spectrum.
	:type top_label: str
	:param bottom_label: string to label the bottom spectrum.
	:type bottom_label: str
	:param xlim: tuple of length 2, defining the beginning and ending values of the x-axis.
	:type xlim: tuple of int
	:param x_threshold: numeric value specifying
	:type x_threshold: float
	:param print_alignment:  whether the intensities should be printed
	:type print_alignment: bool
	:param print_graphic:
	:type print_graphic: bool
	:param output_list: whether the intensities should be returned
	:type output_list: bool
	:return:
	:rtype:
	"""

	if print_graphic:
		import matplotlib.pyplot as plt

	# format spectra and normalize intensitites
	top_tmp = pd.DataFrame(data=spec_top, columns=["mz", "intensity"])
	top_tmp["normalized"] = top_tmp.apply(normalize, args=(max(top_tmp["intensity"]),), axis=1)
	top_tmp = top_tmp[top_tmp["mz"].between(xlim[0], xlim[1])]
	top_plot = top_tmp[["mz", "normalized"]].copy()  # data frame for plotting spectrum
	top_plot.columns = ["mz", "intensity"]
	top = top_plot[top_plot["intensity"] >= b]  # data frame for similarity score calculation

	bottom_tmp = pd.DataFrame(data=spec_bottom, columns=["mz", "intensity"])
	bottom_tmp["normalized"] = bottom_tmp.apply(normalize, args=(max(bottom_tmp["intensity"]),), axis=1)
	bottom_tmp = bottom_tmp[bottom_tmp["mz"].between(xlim[0], xlim[1])]
	bottom_plot = bottom_tmp[["mz", "normalized"]].copy()  # data frame for plotting spectrum
	bottom_plot.columns = ["mz", "intensity"]
	bottom = bottom_plot[bottom_plot["intensity"] >= b]  # data frame for similarity score calculation

	# align the m/z axis of the two spectra, the bottom spectrum is used as the reference

	# Unimplemented R code
	#   for(i in 1:nrow(bottom))
	# 	top["mz"][bottom["mz"][i] >= top["mz"] - t & bottom["mz"][i] <= top["mz"] + t] = bottom["mz"][i]
	# 	top[,1][bottom[,1][i] >= top[,1] - t & bottom[,1][i] <= top[,1] + t] <- bottom[,1][i]
	#   alignment <- merge(top, bottom, by = 1, all = TRUE)
	#   if(length(unique(alignment[,1])) != length(alignment[,1])) warning("the m/z tolerance is set too high")
	# alignment[,c(2,3)][is.na(alignment[,c(2,3)])] <- 0   # convert NAs to zero (R-Help, Sept. 15, 2004, John Fox)
	# names(alignment) <- c("mz", "intensity.top", "intensity.bottom")
	#
	alignment = pd.merge(top, bottom, on="mz", how="outer")
	alignment.fillna(value=0, inplace=True)  # Convert NaN to 0
	alignment.columns = ["mz", "intensity_top", "intensity_bottom"]
	if print_alignment:
		with pd.option_context('display.max_rows', None, 'display.max_columns', None):
			print(alignment)

	# similarity score calculation

	if x_threshold < 0:
		print("Error: x_threshold argument must be zero or a positive number")
		return 1

	# Unimplemented R code
	# alignment <- alignment[alignment[,1] >= x.threshold, ]

	u = np.array(alignment.iloc[:, 1])
	v = np.array(alignment.iloc[:, 2])

	similarity_score = np.dot(u, v) / (np.sqrt(np.sum(np.square(u))) * np.sqrt(np.sum(np.square(v))))

	# Reverse Match
	reverse_alignment = pd.merge(top, bottom, on="mz", how="right")
	reverse_alignment = reverse_alignment.dropna()  # Remove rows containing NaN
	reverse_alignment.columns = ["mz", "intensity_top", "intensity_bottom"]
	u = np.array(reverse_alignment.iloc[:, 1])
	v = np.array(reverse_alignment.iloc[:, 2])

	reverse_similarity_score = np.dot(u, v) / (np.sqrt(np.sum(np.square(u))) * np.sqrt(np.sum(np.square(v))))

	# generate plot

	if print_graphic:
		fig, ax = plt.subplots()
		# fig.scatter(top_plot["mz"],top_plot["intensity"], s=0)
		ax.vlines(top_plot["mz"], 0, top_plot["intensity"], color="blue")
		ax.vlines(bottom["mz"], 0, -bottom["intensity"], color="red")
		ax.set_ylim(-125, 125)
		ax.set_xlim(xlim[0], xlim[1])
		ax.axhline(color="black", linewidth=0.5)
		ax.set_ylabel("Intensity (%)")
		ax.set_xlabel("m/z", style="italic", family="serif")

		h_centre = xlim[0] + (xlim[1] - xlim[0]) // 2

		ax.text(h_centre, 110, top_label, horizontalalignment="center", verticalalignment="center")
		ax.text(h_centre, -110, bottom_label, horizontalalignment="center", verticalalignment="center")
		plt.show()

	# Unimplemented R code
	# 	ticks <- c(-100, -50, 0, 50, 100)
	# 	plot.window(xlim = c(0, 20), ylim = c(-10, 10))
	#
	#
	#   if(output.list == TRUE) {
	#
	# 	# Grid graphics head to tail plot
	#
	# 	headTailPlot <- function() {
	#
	# 	  pushViewport(plotViewport(c(5, 5, 2, 2)))
	# 	  pushViewport(dataViewport(xscale = xlim, yscale = c(-125, 125)))
	#
	# 	  grid.rect()
	# 	  tmp <- pretty(xlim)
	# 	  xlabels <- tmp[tmp >= xlim[1] & tmp <= xlim[2]]
	# 	  grid.xaxis(at = xlabels)
	# 	  grid.yaxis(at = c(-100, -50, 0, 50, 100))
	#
	# 	  grid.segments(top_plot$mz,
	# 					top_plot$intensity,
	# 					top_plot$mz,
	# 					rep(0, length(top_plot$intensity)),
	# 					default.units = "native",
	# 					gp = gpar(lwd = 0.75, col = "blue"))
	#
	# 	  grid.segments(bottom_plot$mz,
	# 					-bottom_plot$intensity,
	# 					bottom_plot$mz,
	# 					rep(0, length(bottom_plot$intensity)),
	# 					default.units = "native",
	# 					gp = gpar(lwd = 0.75, col = "red"))
	#
	# 	  grid.abline(intercept = 0, slope = 0)
	#
	# 	  grid.text("intensity (%)", x = unit(-3.5, "lines"), rot = 90)
	# 	  grid.text("m/z", y = unit(-3.5, "lines"))
	#
	# 	  popViewport(1)
	# 	  pushViewport(dataViewport(xscale = c(0, 20), yscale = c(-10, 10)))
	# 	  grid.text(top.label, unit(10, "native"), unit(9, "native"))
	# 	  grid.text(bottom.label, unit(10, "native"), unit(-9, "native"))
	#
	# 	  popViewport(2)
	#
	# 	}
	#
	# 	p <- grid.grabExpr(headTailPlot())
	#
	#   }
	#
	#
	#

		# with pd.option_context('display.max_rows', None, 'display.max_columns', None):
		# 	print(similarity_score)

	if output_list:
		return similarity_score, reverse_similarity_score, alignment
	# Unimplemented R code
	#
	# return(list(similarity.score = similarity_score,
	# 	alignment = alignment,
	# 	plot = p))
	#
	else:
		return similarity_score, reverse_similarity_score


# simscore <- as.vector((u %*% v)^2 / (sum(u^2) * sum(v^2)))   # cos squared


def normalize(row, max_val):
	# http://jonathansoma.com/lede/foundations/classes/pandas%20columns%20and%20functions/apply-a-function-to-every-row-in-a-pandas-dataframe/
	return (row["intensity"] / float(max_val)) * 100.0


def create_array(intensities, mz):
	return np.column_stack((mz, intensities))



#!/usr/bin/env python3
#
#  constants.py
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from collections import namedtuple

# 3rd party
import quantities

_anions = {  # Incomplete
		'F-': 'fluoride',
		'Cl-': 'chloride',
		'Br-': 'bromide',
		'I-': 'iodide',
		'OH-': 'hydroxide',
		'CN-': 'cyanide',
		'SCN-': 'thiocyanate',
		'CO3-2': 'carbonate',
		'C2O4-2': 'oxalate',
		'HCO3-': 'hydrogencarbonate',
		'NO3-': 'nitrate',
		'NO2-': 'nitrite',
		'PO4-3': 'phospahte',
		'HPO4-2': 'hydrogenphospahte',
		'H2PO4-': 'dihydrogenphospahte',
		'P-3': 'phosphide',
		'SO4-2': 'sulphate',
		'HSO4-': 'hydrogensulphate',
		'SO3-2': 'sulphite',
		'HSO3-': 'hydrogensulphite',
		'S-2': 'sulfide',
		'ClO-': 'hypochlorite',
		'ClO2-': 'chlorite',
		'ClO3-': 'chlorate',
		'ClO4-': 'perchlorate',
		'CrO4-2': 'chromate(VI)',
		'Cr2O7-2': 'dichromate(VI)',
		'MnO4-2': 'manganate(VI)',
		'MnO4-': 'permanganate(VII)',
		'FeO4-2': 'ferrate(VI)',
		'OsO4-2': 'osmate(VI)',
		'Bo3-3': 'borate',
		'BiO3-': 'bismuthate(V)',
		}

_cations = {  # Incomplete
		'H3O+': 'hydronium',
		}

_cation_oxidation_states = {  # This needs to be reviewed, just from the top of my head
		'Cr': (2, 3),
		'Fe': (2, 3),
		'Mn': (2,),
		'Co': (2, 3),
		'Ni': (2, 3),
		'Cu': (1, 2, 3),
		'Ag': (1, 2),
		'Au': (3,),
		'Zn': (2,),
		'Cd': (2,),
		'Hg': (1, 2),  # Tricky: Hg2+2
		'Al': (3,),
		'Ga': (3,),
		'In': (3,),
		'Tl': (1, 3),
		'Sn': (2, 4),
		'Pb': (2, 4),
		'Bi': (3,),
		'Sb': (3,),
		}


class Constant(namedtuple('__BaseConstant', 'name value unit symbol')):
	# TODO: docstring

	# make symbol and unit optional
	def __new__(cls, name: str, value: float, unit: quantities.quantity.Quantity = None, symbol: str = None):
		return super().__new__(cls, name, value, unit, symbol)

	def as_quantity(self):
		"""
		Returns the constant as a :class:`quantities.quantity.Quantity` object.

		:rtype: :class:`quantities.quantity.Quantity`
		"""

		return self.value * self.unit

	def __float__(self):
		return float(self.value)

	def __int__(self):
		return int(self.value)


# The following from periodictable
# public domain data
# Author: Paul Kienzle
avogadro_number = avogadro_constant = Constant(
		name="Avogadro constant",
		value=6.02214179e23,
		unit=1 / quantities.mol,
		symbol="N<sub>A</sub>")  # (30)
plancks_constant = planck_constant = Constant(
		name="Planck's constant",
		value=4.13566733e-15 * (10 ** 34),
		unit=quantities.electron_volt / quantities.second,
		symbol="h")  # (10)
speed_of_light = Constant(
		name="Speed of Light",
		value=299792458,
		unit=quantities.m / quantities.second,
		symbol="c")  # (exact)
electron_radius = Constant(
		name="Electron Radius",
		value=2.8179402894e-15,
		unit=quantities.m,
		symbol="rₑ")  # (58)

# From NIST Reference on Constants, Units, and Uncertainty
#   http://physics.nist.gov/cuu/index.html
# neutron mass = 1.008 664 915 97(43) u
# atomic mass constant m_u = 1.660 538 782(83) x 10-27 kg
neutron_mass = Constant(
		name="Neutron mass",
		value=1.00866491597,
		unit=quantities.atomic_mass_unit,
		symbol="n<sup>o</sup>")  # (43)
atomic_mass_constant = float(quantities.atomic_mass_unit.rescale(quantities.kg))
faraday_constant = Constant(
		name="Faraday constant",
		value=96485.3321233100184,
		unit=quantities.coulomb * (1 / quantities.mol),
		symbol="F")
vacuum_permittivity = Constant(
		"Vacuum permittivity",
		value=8.8541878128e-12,
		unit=quantities.farad / quantities.metre,
		symbol="ε₀")
boltzmann_constant = Constant(
		name="Boltzmann constant",
		value=1.380649e-23,
		unit=quantities.joule / quantities.kelvin,
		symbol="k<sub>B</sub>")
molar_gas_constant = Constant(
		name="Molar gas constant",
		value=8.31446261815324,
		unit=quantities.joule / quantities.kelvin / quantities.mol,
		symbol="R")


# IUPAC prefixes
prefixes = {
		1: "mono",
		2: "di",
		3: "tri",
		4: "tetra",
		5: "penta",
		6: "hexa",
		7: "hepta",
		8: "octa",
		9: "nona",
		10: "deca",
		11: "undeca",
		12: "dodeca",
		13: "trideca",
		14: "tetradeca",
		15: "pentadeca",
		16: "hexadeca",
		17: "heptadeca",
		18: "octadeca",
		19: "nonadeca",
		20: "icosa",
		21: "henicosa",
		22: "docosa",
		23: "tricosa",
		30: "triaconta",
		31: "hentriaconta",
		32: "dotriaconta",
		40: "tetraconta",
		50: "pentaconta",
		60: "hexaconta",
		70: "heptaconta",
		80: "octaconta",
		90: "nonaconta",
		100: "hecta",
		200: "dicta",
		300: "tricta",
		400: "tetracta",
		500: "pentacta",
		600: "hexacta",
		700: "heptacta",
		800: "octacta",
		900: "nonacta",
		1000: "kilia",
		2000: "dilia",
		3000: "trilia",
		4000: "tetralia",
		5000: "pentalia",
		6000: "hexalia",
		7000: "heptalia",
		8000: "octalia",
		9000: "nonalia",
		}



#!/usr/bin/env python3
#
#  formula.py
"""
Parse formulae into a Python object
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |


# stdlib
import math
from collections import Counter, defaultdict
from itertools import combinations_with_replacement, product

#  3rd party
from mathematical.utils import gcd_array

# this package
from chemistry_tools.elements import ELEMENTS, isotope_data, D, T
from chemistry_tools.formulae.parser import string_to_composition
from ._parser_core import _make_isotope_string, _parse_isotope_string
from .iso_dist import IsotopeDistribution
from .utils import _hill_order, _split_isotope, GROUPS
from .composition import Composition


class Formula(defaultdict, Counter):
	"""
	A Formula object stores a chemical composition of a compound.
	It is based on :class:`python:dict`, with the symbols of chemical elements
	as keys and the values equal to the number of atoms of the corresponding
	element in the compound.
	"""
	# TODO: option to convert D and T to H[2] and H[3] ("heavy hydrogen")

	def __init__(self, composition=None, charge=0):
		"""
		:param composition: A :py:class:`Formula` object with the elemental composition of a substance,
			or a :class:`python:dict` representing the same.
			If ``None`` an empty object is created
		:type composition: :py:class:`Formula`, optional
		:param charge:
		:type charge: int, optional
		"""

		defaultdict.__init__(self, int)

		self.charge = 0

		if composition is not None:
			for isotope_string, num_atoms in composition.items():
				element_name, isotope_num = _split_isotope(isotope_string)

				# Remove explicitly undefined isotopes (e.g. X[0]).
				self[_make_isotope_string(element_name, isotope_num)] = num_atoms

			if isinstance(composition, Formula):
				if composition.charge:
					if charge and composition.charge:
						if charge != composition.charge:
							# No point raising error if the charges are the same
							raise ValueError(
									'Charge is specified in both the composition and with the '
									f'`charge` keyword argument. {charge}, {self.charge}')

					charge = composition.charge

		self._set_charge(charge)

	def _set_charge(self, charge: int):
		# Get charge
		if charge and self.charge:
			if charge != self.charge:
				# No point raising error if the charges are the same
				raise ValueError(
						'Charge is specified in both the formula and with the '
						f'`charge` keyword argument. {charge}, {self.charge}')
		elif charge:
			self.charge = charge

	@classmethod
	def from_string(cls, formula, charge=0):
		"""
		Create a new :class:`Formula` object by parsing a string

		.. note:: Isotopes cannot (currently) be parsed using this method

		:param formula: A string with a chemical formula
		:type formula:
		:param charge:
		:type charge: int, optional

		:rtype: :class:`Formula`
		# TODO: should throw error for unrecognised elements CGCGAATTCGCG
		"""

		_class = cls()

		formula = str(formula)
		formula = formula.strip().replace(' ', '')

		# Substitute abbreviations of common chemical groups
		for grp in reversed(sorted(GROUPS)):
			formula = formula.replace(grp, f'({GROUPS[grp]})')

		comp_and_charge = string_to_composition(formula)
		# print(comp_and_charge)

		if 0 in comp_and_charge:
			if charge:
				if comp_and_charge[0] != charge:
					raise ValueError("Cannot supply 'charge' when the formula already has a charge!")

			charge = comp_and_charge[0]

		for symbol, number in comp_and_charge.items():
			if number == 0:
				raise ValueError(f"Unrecognised formula: {formula}")
			# isotope = 0
			if symbol == 0:
				continue

			# isotope_re_1 = re.match(r"([A-z]+)(\[[0-9]*\])", symbol)
			# isotope_re_2 = re.findall(r"(\[[A-z]+)([0-9]*\])", symbol)
			#
			# if isotope_re_1:
			# 	elem, isotope = isotope_re_1[0]
			# 	isotope = re.findall(r"(\[)([0-9]+)(\])", isotope)[0][1]
			# 	if not isotope:
			# 		isotope = 0
			# elif isotope_re_2:
			# 	elem, isotope = isotope_re_2[0]
			# 	elem = elem.lstrip("[")
			# 	isotope = isotope.rstrip("]")
			# elif symbol not in ELEMENTS:
			# 	raise ValueError(f'Unknown chemical element with symbol {symbol}')
			# else:
			# 	elem = ELEMENTS[symbol].symbol
			elem, isotope = _split_isotope(symbol)

			iso_str = _make_isotope_string(elem, int(isotope) if isotope else 0)
			_class[iso_str] += int(number) if number else 1

		_class._set_charge(charge)
		return _class

	@classmethod
	def from_mass_fractions(cls, fractions, charge=0, maxcount=10, precision=1e-4):
		"""
		Create a new :class:`Formula` object from elemental mass fractions by parsing a string

		.. note:: Isotopes cannot (currently) be parsed using this method

		:param fractions: A dictionary of elements and mass fractions
		:type fractions: dict
		:param charge:
		:type charge: int, optional

		:rtype: :class:`Formula`

		Examples
		--------
		>>> Formula.from_mass_fractions({'H': 0.112, 'O': 0.888})
		'H2O'
		>>> Formula.from_mass_fractions({'D': 0.2, 'O': 0.8})
		'O[2H]2'
		>>> Formula.from_mass_fractions({'H': 8.97, 'C': 59.39, 'O': 31.64})
		'C5H9O2'
		>>> Formula.from_mass_fractions({'O': 0.26, '30Si': 0.74})
		'O2[30Si]3'

		"""

		# divide normalized fractions by element/isotope mass
		numbers = {}
		sumfractions = sum(fractions.values())
		for symbol, fraction in fractions.items():
			if symbol[0].isupper():
				try:
					mass = ELEMENTS[symbol].mass
				except KeyError as exc:
					raise ValueError(f"Unknown element '{symbol}'") from exc
			else:
				if symbol.startswith('['):
					symbol = symbol[1:-1]
				i = 0
				while symbol[i].isdigit():
					i += 1
				massnum = int(symbol[:i])
				symbol = symbol[i:]
				try:
					mass = ELEMENTS[symbol].isotopes[massnum].mass
				except KeyError as exc:
					raise ValueError(f"Unknown isotope '[{massnum}{symbol}]'") from exc
				symbol = f'[{massnum}{symbol}]'
			numbers[symbol] = fraction / (sumfractions * mass)

		# divide numbers by smallest number
		smallest = min(numbers.values())
		for symbol in numbers:
			numbers[symbol] /= smallest

		# find smallest factor that turns all numbers into integers
		precision *= len(numbers)
		best = 1e6
		factor = 1
		for i in range(1, maxcount):
			x = sum(abs((i * n) - round(i * n)) for n in numbers.values())
			if x < best:
				best = x
				factor = i
				if best < i * precision:
					break

		formula = []
		for symbol, number in sorted(numbers.items()):
			count = int(round(factor * number))
			if count > 1:
				formula.append(f'{symbol}{count}')
			else:
				formula.append(symbol)

		return cls.from_string(''.join(formula), charge=charge)

	@classmethod
	def from_kwargs(cls, *, charge=0, **kwargs):
		"""
		Create a new :class:`Formula` object from keyword arguments representing the elements in the compound

		:param charge:
		:type charge: int, optional

		:rtype: :class:`Formula`
		"""

		return cls(kwargs, charge=charge)

	@property
	def monoisotopic_mass(self):
		"""
		Calculate the monoisotopic mass of a :py:class:`Formula`.
		If any isotopes are already present in the formula, the mass of these will be preserved

		:return: mass
		:rtype: float
		"""

		# Calculate mass
		mass = 0.0

		for element, count in self.items():
			if element == "D":
				iso_mass = D.mass
			elif element == "T":
				iso_mass = T.mass
			else:
				symbol, isotope = _split_isotope(element)

				if isotope:
					iso_mass = ELEMENTS[symbol].isotopes[isotope].mass
				else:
					iso_mass = ELEMENTS[symbol].isotopes[ELEMENTS[symbol].nominalmass].mass

			mass += iso_mass * count

		return mass

	nominal_mass = exact_mass = monoisotopic_mass

	@property
	def mass(self):
		"""
		Calculate the average mass of a :py:class:`Formula`.

		Note that mass is not averaged for elements with specified isotopes.

		:return: mass
		:rtype: float
		"""

		# Calculate mass
		mass = 0.0

		for element, count in self.items():
			if element == "D":
				iso_mass = D.mass
			elif element == "T":
				iso_mass = T.mass
			else:
				symbol, isotope = _split_isotope(element)

				if isotope:
					iso_mass = ELEMENTS[symbol].isotopes[isotope].mass
				else:
					iso_mass = ELEMENTS[symbol].mass

			mass += iso_mass * count

		return mass

	average_mass = mass

	@property
	def mz(self):
		return self.get_mz()

	@property
	def average_mz(self):
		return self.get_mz(average=True)

	def get_mz(self, average=True, charge=None):
		"""
		Calculate the average mass:charge ratio (*m/z*) of a :py:class:`Formula`.

		:param average: If :py:const:`True` then the average *m/z* is calculated. Note that mass
			is not averaged for elements with specified isotopes. Default is
			:py:const:`True`.
		:type: average: bool, optional
		:param charge: The charge of the compound. If ``None`` then the existing charge of the Formula is used
		:type: charge: int, optional

		:return: mass
		:rtype: float
		"""

		if average:
			mass = self.average_mass
		else:
			mass = self.exact_mass

		# Calculate m/z
		if charge:
			mass /= charge
		elif self.charge:
			mass /= self.charge

		return mass

	def most_probable_isotopic_composition(self, elements_with_isotopes=None):
		"""
		Calculate the most probable isotopic composition of a molecule/ion.

		For each element, only two most abundant isotopes are considered.
		Any isotopes already in the Formula will be changed to the most abundant isotope

		:param elements_with_isotopes: A set of elements to be considered in isotopic distribution
			(by default, every element has an isotopic distribution).
		:type: elements_with_isotopes: container of str, optional
		:return: A tuple with the most probable isotopic composition and its
			relative abundance.
		:rtype: (Formula, float)
		"""

		# Removing isotopes from the composition.
		for isotope_string in self:
			element_name, isotope_num = _split_isotope(isotope_string)
			if isotope_num:
				self[element_name] += self.pop(isotope_string)

		isotopic_composition = Formula()

		for element_name in self:
			if not elements_with_isotopes or (element_name in elements_with_isotopes):
				# Take the two most abundant isotopes.
				first_iso, second_iso = sorted(
						[(i[0], i[1][1]) for i in isotope_data[element_name].items() if i[0]],
						key=lambda x: -x[1]
						)[:2]

				# Write the number of isotopes of the most abundant type.
				first_iso_str = _make_isotope_string(element_name, first_iso[0])
				isotopic_composition[first_iso_str] = int(math.ceil(self[element_name])) * first_iso[1]

				# Write the number of the second isotopes.
				second_iso_str = _make_isotope_string(element_name, second_iso[0])
				isotopic_composition[second_iso_str] = self[element_name] - isotopic_composition[first_iso_str]
			else:
				isotopic_composition[element_name] = self[element_name]

		return isotopic_composition, isotopic_composition.isotopic_composition_abundance

	@property
	def isotopic_composition_abundance(self):
		"""
		Calculate the relative abundance of the current isotopic composition
		of this molecule.

		:returns: The relative abundance of the current isotopic composition.
		:rtype: float
		"""

		isotopic_composition = defaultdict(dict)

		# Check if there are default and non-default isotopes of the same
		# element and rearrange the elements.
		for element in self:
			element_name, isotope_num = _split_isotope(element)

			# If there is already an entry for this element and either it
			# contains a default isotope or newly added isotope is default
			# then raise an exception.
			if (
					(element_name in isotopic_composition)
					and (isotope_num == 0 or 0 in isotopic_composition[element_name])
					):
				raise ValueError(
						'Please specify the isotopic states of all atoms of '
						f'{element_name} or do not specify them at all.'
						)
			else:
				isotopic_composition[element_name][isotope_num] = (
						self[element])

		# Calculate relative abundance.
		num1, num2, denom = 1, 1, 1

		for element_name, isotope_dict in isotopic_composition.items():
			num1 *= math.factorial(sum(isotope_dict.values()))
			for isotope_num, isotope_content in isotope_dict.items():
				denom *= math.factorial(isotope_content)
				if isotope_num:
					num2 *= (isotope_data[element_name][isotope_num][1] ** isotope_content)

		return num2 * (num1 / denom)

	def iter_isotopologues(
			self, report_abundance=False, elements_with_isotopes=None,
			isotope_threshold=5e-4, overall_threshold=0):
		"""
		Iterate over possible isotopic states of the molecule.

		The space of possible isotopic compositions is restrained by parameters
		``elements_with_isotopes``, ``isotope_threshold``, ``overall_threshold``.

		:param report_abundance: If :py:const:`True`, the output will contain 2-tuples: `(composition, abundance)`.
			Otherwise, only compositions are yielded. Default is :py:const:`False`.
		:type: report_abundance: bool, optional
		:param elements_with_isotopes: A set of elements to be considered in isotopic distribution
			(by default, every element has an isotopic distribution).
		:type: elements_with_isotopes: container of str, optional
		:param isotope_threshold: The threshold abundance of a specific isotope to be considered.
			Default is :py:const:`5e-4`.
		:type: isotope_threshold: float, optional
		:param overall_threshold: The threshold abundance of the calculated isotopic composition.
			Default is :py:const:`0`.
		:type: overall_threshold: float, optional

		:return: Iterator over possible isotopic compositions.
		:rtype: iterator
		"""

		dict_elem_isotopes = {}
		for element in self:
			if elements_with_isotopes is None or element in elements_with_isotopes:
				element_name, isotope_num = _split_isotope(element)
				isotopes = {k: v for k, v in isotope_data[element_name].items() if k != 0 and v[1] >= isotope_threshold}
				list_isotopes = [_make_isotope_string(element_name, k) for k in isotopes]
				dict_elem_isotopes[element] = list_isotopes
			else:
				dict_elem_isotopes[element] = [element]

		all_isotoplogues = []
		for element, list_isotopes in dict_elem_isotopes.items():
			n = self[element]
			list_comb_element_n = []
			for elementXn in combinations_with_replacement(list_isotopes, n):
				list_comb_element_n.append(elementXn)
			all_isotoplogues.append(list_comb_element_n)

		for isotopologue in product(*all_isotoplogues):
			flat_isotopologue = [atom for element in isotopologue for atom in element]
			ic = Formula(Counter(flat_isotopologue))
			if report_abundance or overall_threshold > 0.0:
				abundance = ic.isotopic_composition_abundance
				if abundance > overall_threshold:
					if report_abundance:
						yield ic, abundance
					else:
						yield ic
			else:
				yield ic

	def isotope_distribution(self):
		"""
		Returns a :class:`IsotopeDistribution` object representing the distribution of the
		isotopologues of the formula

		:rtype: :class:`IsotopeDistribution`
		"""

		return IsotopeDistribution(self)

	def copy(self):
		return self.__class__(self, charge=self.charge)

	def __reduce__(self):
		class_, args, state, list_iterator, dict_iterator = super(
				defaultdict, self).__reduce__()
		# Override the reduce of defaultdict so we do not provide the
		# `int` type as the first argument
		# which prevents from correctly unpickling the object
		args = ()
		return class_, args, state, list_iterator, dict_iterator

	def __missing__(self, key):
		# override default behavior: we don't want to add 0's to the dictionary
		return 0

	def __setitem__(self, key, value):
		if isinstance(value, float):
			value = int(round(value))
		elif not isinstance(value, int):
			raise TypeError(f'Only integers allowed as values in Formula, got {type(value).__name__}.')
		if value:  # reject 0's
			super(defaultdict, self).__setitem__(key, value)
		elif key in self:
			del self[key]

	def __add__(self, other):
		result = self.copy()
		for elem, count in other.items():
			result[elem] += count
		return result

	def __iadd__(self, other):
		for elem, count in other.items():
			self[elem] += count
		return self

	def __radd__(self, other):
		return self + other

	def __sub__(self, other):
		result = self.copy()
		for elem, count in other.items():
			result[elem] -= count
		return result

	def __isub__(self, other):
		for elem, count in other.items():
			self[elem] -= count
		return self

	def __rsub__(self, other):
		return (self - other) * (-1)

	def __mul__(self, other):
		if not isinstance(other, int):
			raise TypeError(f'Cannot multiply Formula by non-integer "{other}"')
		return type(self)({k: v * other for k, v in self.items()})

	def __imul__(self, other):
		if not isinstance(other, int):
			raise TypeError(f'Cannot multiply Formula by non-integer "{other}"')
		for elem in self:
			self[elem] *= other
		return self

	def __rmul__(self, other):
		return self * other

	def __eq__(self, other):
		if isinstance(other, (dict, Formula)):
			self_items = {i for i in self.items() if i[1]}
			other_items = {i for i in other.items() if i[1]}

			if isinstance(other, Formula):
				return self_items == other_items and self.charge == other.charge
			else:
				return self_items == other_items
		else:
			return NotImplemented

	def __str__(self):
		return f'{type(self).__name__}({", ".join(self._repr_elements())})'

	def _repr_elements(self):
		elements = [str(dict.__repr__(self))]

		if self.charge:
			elements.append(f"charge={self.charge}")

		return elements

	def __repr__(self):
		return str(self)

	def _repr_pretty_(self, p, cycle):
		if cycle:  # should never happen
			p.text('{} object with a cyclic reference'.format(type(self).__name__))
		p.text(str(self))

	@property
	def hill_formula(self):
		"""
		Returns formula in Hill notation

		Examples
		--------
		>>> Formula.from_string('BrC2H5').hill_formula
		'C2H5Br'
		>>> Formula.from_string('HBr').hill_formula
		'BrH'
		>>> Formula.from_string('[(CH3)3Si2]2NNa').hill_formula
		'C6H18NNaSi4'

		:rtype: str
		"""

		hill = []

		for symbol in _hill_order(self.elements):
			hill.append(symbol)
			count = self[symbol]
			if count > 1:
				hill.append(str(count))

		# alphabet = sorted(ELEMENTS.symbols)
		#
		# if "C" in self:
		# 	hill.append("C")
		# 	alphabet.remove("C")
		# 	count = self["C"]
		# 	if count > 1:
		# 		hill.append(str(count))
		# 	if "H" in self:
		# 		hill.append("H")
		# 		alphabet.remove("H")
		# 		count = self["H"]
		# 		if count > 1:
		# 			hill.append(str(count))
		#
		# for element in alphabet:
		# 	if element in self:
		# 		hill.append(element)
		# 		count = self[element]
		# 		if count > 1:
		# 			hill.append(str(count))
		#
		return ''.join(hill)

	@property
	def no_isotope_hill_formula(self):
		"""
		Returns formula in Hill notation, without any isotopes specified

		Examples
		--------
		>>> Formula.from_string('BrC2H5').no_isotope_hill_formula
		'C2H5Br'
		>>> Formula.from_string('HBr').no_isotope_hill_formula
		'BrH'
		>>> Formula.from_string('[(CH3)3Si2]2NNa').no_isotope_hill_formula
		'C6H18NNaSi4'

		:rtype: str
		"""

		hill = []

		ordered_symbols = dict()

		for symbol in _hill_order(self.elements):
			count = self[symbol]
			symbol, _ = _split_isotope(symbol)
			if symbol in ordered_symbols:
				ordered_symbols[symbol] += count
			else:
				ordered_symbols[symbol] = count

		for symbol, count in ordered_symbols.items():
			hill.append(symbol)
			if count > 1:
				hill.append(str(count))

		return ''.join(hill)

	@property
	def empirical_formula(self):
		"""
		Returns the empirical formula in Hill notation.

		The empirical formula has the simplest whole number ratio of atoms
		of each element present in formula.

		Examples
		--------
		>>> Formula.from_string('H2O').empirical
		'H2O'
		>>> Formula.from_string('S4').empirical
		'S'
		>>> Formula.from_string('C6H12O6').empirical
		'CH2O'

		:rtype: str
		"""

		hill = []

		divisor = gcd_array(list(self.values()))

		for symbol in _hill_order(self.elements):
			hill.append(symbol)
			count = self[symbol] // divisor
			if count > 1:
				hill.append(str(count))

		return ''.join(hill)

	@property
	def n_atoms(self):
		"""
		Return the number of atoms in the formula.

		Examples
		--------
		>>> Formula.from_string('CH3COOH').n_atoms
		8

		"""

		return sum(list(self.values()))

	@property
	def n_elements(self):
		"""
		Return the number of elements in the formula.

		Examples
		--------
		>>> Formula.from_string('CH3COOH').n_elements
		3

		"""

		return len(self)

	@property
	def elements(self):
		"""
		Returns a list of the element symbols in the formula.
		"""

		return list(self.keys())

	@property
	def composition(self):
		"""
		Returns a :class:`Composition` object representing the elemental composition of the Formula

		:return:
		:rtype:
		"""

		return Composition(self)



#!/usr/bin/env python3
#
#  dataarray.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from abc import abstractmethod

#  3rd party
import pandas
import tabulate
from cawdrey import FrozenOrderedDict


class DataArray(FrozenOrderedDict):
	"""
	A class that can output data as an array, to CSV, as a pandas DataFrame,
	or as a pretty-printed table in a variety of formats.
	"""

	def __init__(self, formula, data):
		"""

		:param formula: A :class:`Formula` object to create the distribution for
		:type formula: :class:`Formula`
		:param data: A dictionary of data to add to the internal
			:class:`~cawdrey.frozenordereddict.FrozenOrderedDict`
		:type data: dict
		"""

		super().__init__(**data)
		self.formula = formula

	def as_csv(self, *args, sep=",", **kwargs):
		"""
		Returns the data as a CSV formatted string

		:param sep: The separator for the CSV data. Default ``,``
		:type sep: str, optional

		Any additional arguments taken by ``as_array`` can also be used here.

		:rtype: str
		"""

		return "\n".join(sep.join(x) for x in self.as_array(*args, **kwargs))

	_as_array_kwargs = set()

	@abstractmethod
	def as_array(self, *args, **kwargs):
		pass

	def as_dataframe(self, *args, **kwargs):
		"""
		Returns the isotope distribution data as a pandas DataFrame

		Any arguments taken by ``as_array`` can also be used here.

		:rtype: :class:`pandas.DataFrame`
		"""

		array = self.as_array(*args, **kwargs)
		return pandas.DataFrame.from_records(array[1:], columns=array[0])

	_as_table_alignment = []
	_as_table_float_format = []

	def as_table(self, *args, **kwargs):
		"""
		Returns the isotope distribution data as a table using
		`tabulate <https://github.com/astanin/python-tabulate>`_

		Any arguments taken by ``as_array`` can also be used here.

		Additionally, any valid keyword argument for :func:`tabulate.tabulate` can be used.

		:rtype: str
		"""

		tabulate_kwargs = {}
		array_kwargs = {}
		for arg, val in kwargs.items():
			if arg in self._as_array_kwargs:
				array_kwargs[arg] = val
			else:
				tabulate_kwargs[arg] = val

		if "colalign" not in tabulate_kwargs:
			tabulate_kwargs["colalign"] = self._as_table_alignment
		if "floatfmt" not in tabulate_kwargs:
			tabulate_kwargs["floatfmt"] = self._as_table_float_format

		array = self.as_array(*args, **array_kwargs)
		return tabulate.tabulate(array[1:], array[0], **tabulate_kwargs)

	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		return f"<{self.__class__.__name__}({self.formula})>"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  _parser_core.py
"""
Core functions and constants for parsing formulae
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


# stdlib
import re
import warnings


_greek_letters = (
		'alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta', 'eta', 'theta',
		'iota', 'kappa', 'lambda', 'mu', 'nu', 'xi', 'omicron', 'pi', 'rho',
		'sigma', 'tau', 'upsilon', 'phi', 'chi', 'psi', 'omega'
		)
_greek_u = 'αβγδεζηθικλμνξοπρστυφχψω'


def _formula_to_format(
		sub, sup, formula, prefixes=None,
		infixes=None, suffixes=('(s)', '(l)', '(g)', '(aq)')):
	# TODO: make isootope square brackets be superscript
	parts = _formula_to_parts(formula, prefixes.keys(), suffixes)
	stoichs = parts[0].split('.')
	string = ''

	for idx, stoich in enumerate(stoichs):
		if idx == 0:
			m = 1
		else:
			m, stoich = _get_leading_integer(stoich)
			string += _subs('.', infixes)
		if m != 1:
			string += str(m)

		string += re.sub(r'([0-9]+)(?![^\[]*\])', lambda m: sub(m.group(1)), stoich)

	if parts[1] is not None:
		chg = _get_charge(parts[1])
		if chg < 0:
			token = '-' if chg == -1 else '%d-' % -chg
		if chg > 0:
			token = '+' if chg == 1 else '%d+' % chg
		string += sup(token)
	if len(parts) > 4:
		raise ValueError("Incorrect formula")
	pre_str = ''.join(map(lambda x: _subs(x, prefixes), parts[2]))
	return pre_str + string + ''.join(parts[3])


def _formula_to_parts(formula, prefixes, suffixes):
	# Drop prefixes and suffixes
	drop_pref, drop_suff = [], []
	for ign in prefixes:
		if formula.startswith(ign):
			drop_pref.append(ign)
			formula = formula[len(ign):]
	for ign in suffixes:
		if formula.endswith(ign):
			drop_suff.append(ign)
			formula = formula[:-len(ign)]

	# Extract charge
	for token in '+-':
		if token in formula:
			if formula.count(token) > 1:
				raise ValueError("Multiple tokens: %s" % token)
			parts = formula.split(token)
			parts[1] = token + parts[1]
			break
	else:
		parts = [formula, None]
	return parts + [tuple(drop_pref), tuple(drop_suff[::-1])]


def _subs(string, patterns):
	for patt, repl in patterns.items():
		string = string.replace(patt, repl)
	return string


def _get_leading_integer(s):
	m = re.findall(r'^\d+', s)
	if len(m) == 0:
		m = 1
	elif len(m) == 1:
		s = s[len(m[0]):]
		m = int(m[0])
	else:
		raise ValueError("Failed to parse: %s" % s)
	return m, s


def _get_charge(charge_str):
	if charge_str == '+':
		return 1
	elif charge_str == '-':
		return -1

	for token, anti, sign in zip('+-', '-+', (1, -1)):
		if token in charge_str:
			if anti in charge_str:
				raise ValueError("Invalid charge description (+ & - present)")
			before, after = charge_str.split(token)
			if len(before) > 0 and len(after) > 0:
				raise ValueError("Values both before and after charge token")
			if len(before) > 0:
				# will_be_missing_in='0.8.0'
				warnings.warn("'Fe/3+' deprecated, use e.g. 'Fe+3'", DeprecationWarning, stacklevel=3)
				return sign * int(1 if before == '' else before)
			if len(after) > 0:
				return sign * int(1 if after == '' else after)
	raise ValueError("Invalid charge description (+ or - missing)")


def _make_isotope_string(element_name, isotope_num):
	"""
	Form a string label for an isotope.
	"""

	if isotope_num == 0:
		return element_name
	else:
		return f'[{isotope_num}{element_name}]'


_isotope_string = r'^([A-Z][a-z+]*)(?:\[(\d+)\])?$'


# TODO: merge with _split_isotope
def _parse_isotope_string(label):
	"""
	Parse an string with an isotope label and return the element name and
	the isotope number.

	>>> _parse_isotope_string('C')
	('C', 0)
	>>> _parse_isotope_string('C[12]')
	('C', 12)
	"""

	element_name, num = re.match(_isotope_string, label).groups()
	isotope_num = int(num) if num else 0
	return element_name, isotope_num


def _parse_multiplicity(strings, substance_keys=None):
	"""
	Examples
	--------
	>>> _parse_multiplicity(['2 H2O2', 'O2']) == {'H2O2': 2, 'O2': 1}
	True
	>>> _parse_multiplicity(['2 * H2O2', 'O2']) == {'H2O2': 2, 'O2': 1}
	True
	>>> _parse_multiplicity(['']) == {}
	True
	>>> _parse_multiplicity(['H2O', 'H2O']) == {'H2O': 2}
	True

	"""
	result = {}
	for items in [re.split(' \\* | ', s) for s in strings]:
		items = [x for x in items if x != '']
		if len(items) == 0:
			continue
		elif len(items) == 1:
			if items[0] not in result:
				result[items[0]] = 0
			result[items[0]] += 1
		elif len(items) == 2:
			if items[1] not in result:
				result[items[1]] = 0
			result[items[1]] += float(items[0]) if '.' in items[0] or 'e' in items[0] else int(items[0])
		else:
			raise ValueError("To many parts in substring")
	if substance_keys is not None:
		for k in result:
			if k not in substance_keys:
				raise ValueError("Unkown substance_key: %s" % k)
	return result



#!/usr/bin/env python3
#
#  iso_dist.py
"""
Isotope Distributions
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


# stdlib
from collections import OrderedDict
from enum import Enum

# this package
from .dataarray import DataArray
from .unicode import string_to_unicode


class IsoDistSort(Enum):
	"""
	Lookup for sorting isotope distribution output
	"""
	formula = 0
	mass = 1
	abundance = 2
	relative_abundance = 3
	Formula = 0
	Mass = 1
	Abundance = 2
	Relative_abundance = 3
	Relative_Abundance = 3

	def __int__(self):
		return int(self.value)


class IsotopeDistribution(DataArray):
	"""


	Each composition can be accessed with their hill formulae like a dictionary
	(e.g. iso_dict["H[1]2O[16]"]
	"""

	def __init__(self, formula):
		"""

		:param formula: A :class:`Formula` object to create the distribution for
		:type formula: :class:`Formula`
		"""

		iso_compositions = list(formula.iter_isotopologues())
		compositions = OrderedDict()
		max_abundance = 0

		for comp in iso_compositions:
			compositions[comp.hill_formula] = comp
			abundance = comp.isotopic_composition_abundance
			if abundance > max_abundance:
				max_abundance = abundance

		super().__init__(formula=iso_compositions[0].no_isotope_hill_formula, data=compositions)

		self.max_abundance = max_abundance

	_as_array_kwargs = {"sort_by", "reverse", "format_percentage"}
	_as_table_alignment = ["left", "right", "right", "right"]
	_as_table_float_format = [None, ".4f", ".6f", ".6f"]

	def as_array(self, sort_by=IsoDistSort.formula, reverse=False, format_percentage=True):
		"""
		Returns the isotope distribution data as a list of lists

		:param sort_by: The column to sort by.
		:type sort_by: IsoDistSort
		:param: Whether the isotopologues should be sorted in reverse order. Default ``False``.
		:type reverse: bool, optional
		:param: Whether the abundances should be formatted as percentages or not. Default ``False``.
		:type format_percentage: bool, optional

		:rtype: list[list]
		"""

		if sort_by not in IsoDistSort:
			raise ValueError(f"Unrecognised value for 'sort_by': {sort_by}")
		elif sort_by == IsoDistSort.formula:
			sort_key = lambda comp: comp.hill_formula
		elif sort_by == IsoDistSort.mass:
			sort_key = lambda comp: comp.mass
		elif sort_by == IsoDistSort.abundance:
			sort_key = lambda comp: comp.isotopic_composition_abundance
		elif sort_by == IsoDistSort.relative_abundance:
			sort_key = lambda comp: comp.isotopic_composition_abundance / self.max_abundance
		else:
			raise ValueError(f"Unrecognised value for 'sort_by': {sort_by}")

		output = []

		for comp in sorted(self.values(), key=sort_key, reverse=reverse):
			abundance = comp.isotopic_composition_abundance
			rel_abund = abundance / self.max_abundance
			row = [comp.hill_formula, f"{comp.mass:0.4f}"]
			if format_percentage:
				row += [f"{abundance:0.2%}", f"{rel_abund:0.2%}"]
			else:
				row += [f"{abundance:0.6f}", f"{rel_abund:0.6f}"]
			output.append(row)
			# TODO: Unicode, latex, html representations of formulae

		output.insert(0, ["Formula", "Mass", "Abundance", "Relative Abundance"])

		return output

	def __str__(self):
		table = self.as_table(sort_by=IsoDistSort.relative_abundance, reverse=True, tablefmt="fancy_grid")
		return f"\n Isotope Distribution for {string_to_unicode(self.formula)}\n{table}"

# TODO: as_mass_spec



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  pug_rest.py
"""
Functions for interacting with PubChem PUG_REST API
"""
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
import time
from urllib.parse import quote

# this package
from chemistry_tools import cached_requests
from chemistry_tools.pubchem import API_BASE
from chemistry_tools.pubchem.enums import PubChemFormats, PubChemNamespace
from chemistry_tools.pubchem.errors import HTTP_ERROR_CODES, PubChemHTTPError
from chemistry_tools.pubchem.utils import _force_sequence_or_csv, _make_base_url


def _do_rest_get(
		namespace, identifier, format_=PubChemFormats.JSON, domain=None,
		record_type="2d",
		png_width=300, png_height=300):
	"""
	:param identifier: Identifiers (e.g. name, CID) for the compounds to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:type identifier: str, Sequence[str]
	:param namespace: The type of identifier to look up. Valid values are in :class:`PubChemNamespace`. Default "name"
	:type namespace: PubChemNamespace, optional
	:type namespace: str, optional
	:param format_: The file format to retrieve the data in. Valid values are in :class:`PubChemFormats`, plus "PNG"
	:type format_: PubChemFormats or str, optional

	:return:
	:rtype:
	"""

	# domain = description, synonyms, or property followed by a comma-separated list of desired properties

	if not PubChemNamespace.is_valid_value(namespace):
		raise ValueError(f"'{namespace}' is not a valid value for 'namespace'")

	if not PubChemFormats.is_valid_value(format_):
		raise ValueError(f"'{format_}' is not a valid value for 'format_'")

	query_params = {}

	if str(format_).upper() == str(PubChemFormats.PNG):
		query_params["image_size"] = f"{png_width}x{png_height}"

	if domain:
		r = cached_requests.get(f"{_make_base_url(namespace, identifier)}/{domain}/{format_}", params=query_params)
	else:
		query_params["record_type"] = record_type
		r = cached_requests.get(f"{_make_base_url(namespace, identifier)}/{format_}", params=query_params)

	if r.status_code in HTTP_ERROR_CODES:
		raise PubChemHTTPError(r)

	return r


def get_full_json(cid):
	json_file = cached_requests.get(f"{API_BASE}_view/data/compound/{cid}/JSON/")
	return json_file.json()


def async_get(identifier, namespace='cid', operation=None, output='JSON', searchtype=None, **kwargs):
	"""
	Request wrapper that automatically handles async requests.
	"""
	# TODO:

	if (searchtype and searchtype != 'xref') or namespace in ['formula']:
		r = request(identifier, namespace, None, 'JSON', searchtype, **kwargs)
		response = r.content
		status = r.json()
		if 'Waiting' in status and 'ListKey' in status['Waiting']:
			identifier = status['Waiting']['ListKey']
			namespace = 'listkey'
			while 'Waiting' in status and 'ListKey' in status['Waiting']:
				time.sleep(2)
				r = request(identifier, namespace, operation, 'JSON', **kwargs)
				response = r.content
				status = r.json()
			if output != 'JSON':
				response = request(identifier, namespace, operation, output, searchtype, **kwargs).content
	else:
		response = request(identifier, namespace, operation, output, searchtype, **kwargs).content

	return response


def request(identifier, namespace='cid', operation=None, output='JSON', searchtype=None, **kwargs):
	"""
	Construct API request from parameters and return the response.

	Full specification at http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST.html
	"""

	# If identifier is a list, join with commas into string
	if isinstance(identifier, int):
		identifier = str(identifier)

	identifier = _force_sequence_or_csv(identifier, "identifier")
	identifier = ','.join(str(x) for x in identifier)

	# Build API URL
	urlid, params = None, {}

	# use this function when:
		# namespace in ['listkey', 'formula']
		# searchtype == 'xref'

	urlid = quote(identifier.encode('utf8'))

	comps = filter(None, [API_BASE, "compound", searchtype, namespace, urlid, operation, output])
	apiurl = '/'.join(comps)

	# Filter None values from kwargs
	for key, val in kwargs.items():
		if val is not None:
			params[key] = val

	# print(f'Request URL: {apiurl}')
	# print(f'Request data: {params}')

	response = cached_requests.get(apiurl, params=params)
	if response.status_code in HTTP_ERROR_CODES:
		raise PubChemHTTPError(response)

	return response



#!/usr/bin/env python3
#
#  compound.py
"""
Parse formulae into a Python object
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#


# this package
import quantities
from domdf_python_tools.bases import Dictable

from chemistry_tools.formulae.formula import Formula
from chemistry_tools.formulae.html import string_to_html
from chemistry_tools.formulae.latex import string_to_latex
from chemistry_tools.formulae.unicode import string_to_unicode


class Compound(Dictable):
	"""
	Class representing a chemical compound.

	Parameters
	----------
	name : str
	latex_name : str
	unicode_name : str
	html_name : str
	formula: Formula
		The chemical formula of the compound. If none this is generated from the name
	data : dict
		Free form dictionary. Could be simple such as ``{'mp': 0, 'bp': 100}``
		or considerably more involved, e.g.: ``{'diffusion_coefficient': {\
		'water': lambda T: 2.1*m**2/s/K*(T - 273.15*K)}}``.

	Attributes
	----------
	mass
		The mass of the compound
	formula:
		The :class:`Formula` object representing the compound
	data
		Free form dictionary of additional properties.
	"""

	def __init__(
			self, name, formula=None, data=None,
			latex_name=None, unicode_name=None, html_name=None,
			):

		super().__init__()

		self.name = name

		if formula is None:
			formula = name
			self.formula = Formula.from_string(formula)
		else:
			self.formula = formula

		if latex_name:
			self.latex_name = latex_name
		else:
			self.latex_name = string_to_latex(self.formula.hill_formula)

		if unicode_name:
			self.unicode_name = unicode_name
		else:
			self.unicode_name = string_to_unicode(self.formula.hill_formula)

		if html_name:
			self.html_name = html_name
		else:
			self.html_name = string_to_html(self.formula.hill_formula)

		self.data = data or {}

	@property
	def __dict__(self):
		return dict(
				name=self.name,
				latex_name=self.latex_name,
				unicode_name=self.unicode_name,
				html_name=self.html_name,
				formula=self.formula,
				data=self.data,
				)

	def __eq__(self, other):
		if isinstance(other, str):
			return self.name == other
		else:
			super().__eq__(other)

	@property
	def charge(self):
		"""
		Convenience prop for accessing charge of the formula
		"""

		return self.formula.charge

	@property
	def mass(self):
		"""
		Convenience prop for accessing the mass of the formula
		"""

		return self.formula.mass

	def molar_mass(self):
		"""
		Returns the molar mass (with units) of the substance

		Examples
		--------
		>>> nh4p = Compound.from_formula('NH4+')
		>>> from chemistry_tools.units import quantities
		>>> nh4p.molar_mass(quantities)
		array(18.0384511...) * g/mol

		"""

		return self.mass * quantities.g / quantities.mol

	def __repr__(self):
		return f"<{self.__class__.__name__}({self.name}, {self.formula})>"

	def __str__(self):
		return str(self.name)

	def _repr_html_(self):
		return self.html_name
